@@ -29,11 +29,13 @@
 #include <stdlib.h>
 #include <ctype.h>
 #include "atacmds.h"
-#include "escalade.h"
 #include "extern.h"
 #include "utility.h"
 
-const char *atacmds_c_cvsid="$Id: atacmds.c,v 1.115 2003/08/13 12:33:22 ballen4705 Exp $" ATACMDS_H_CVSID ESCALADE_H_CVSID EXTERN_H_CVSID UTILITY_H_CVSID;
+const char *atacmds_c_cvsid="$Id: atacmds.c,v 1.132 2003/11/28 23:21:58 ballen4705 Exp $" ATACMDS_H_CVSID EXTERN_H_CVSID UTILITY_H_CVSID;
+
+// to hold onto exit code for atexit routine
+extern int exitstatus;
 
 // for passing global control variables
 extern smartmonctrl *con;
@@ -180,7 +182,7 @@ int parse_attribute_def(char *pair, unsigned char **defsptr){
   // If array does not exist, allocate it
   if (!*defsptr && !(*defsptr=(unsigned char *)calloc(MAX_ATTRIBUTE_NUM, 1))){
     pout("Out of memory in parse_attribute_def\n");
-    exit(1);
+    EXIT(1);
   }
 
   defs=*defsptr;
@@ -312,7 +314,7 @@ char *pad_numbers(const char *s)
   r = s + strlen(s);
   u = t;
   while (( r-- >= s)) {
-    if (isdigit(*r))
+    if (isdigit((int)*r))
       ndigits++;
     else if (ndigits > 0) {
       while (ndigits++ < 3)
@@ -473,165 +475,19 @@ void swap8(char *location){
   return;
 }
 
-// PURPOSE
-//   This is an interface routine meant to isolate the OS dependent
-//   parts of the code, and to provide a debugging interface.  Each
-//   different port and OS needs to provide it's own interface.  This
-//   is the linux one.
-// DETAILED DESCRIPTION OF ARGUMENTS
-//   device: is the file descriptor provided by open()
-//   command: defines the different operations.
-//   select: additional input data if needed (which log, which type of
-//           self-test).
-//   data:   location to write output data, if needed (512 bytes).
-//   Note: not all commands use all arguments.
-// RETURN VALUES
-//  -1 if the command failed
-//   0 if the command succeeded,
-//   STATUS_CHECK routine: 
-//  -1 if the command failed
-//   0 if the command succeeded and disk SMART status is "OK"
-//   1 if the command succeeded and disk SMART status is "FAILING"
-
-
-// huge value of buffer size needed because HDIO_DRIVE_CMD assumes
-// that buff[3] is the data size.  Since the SMART_AUTOSAVE and
-// SMART_AUTO_OFFLINE use values of 0xf1 and 0xf8 we need the space.
-// Otherwise a 4+512 byte buffer would be enough.
-#define STRANGE_BUFFER_LENGTH (4+512*0xf8)
-
-int linux_ata_command_interface(int device, smart_command_set command, int select, char *data){
-  unsigned char buff[STRANGE_BUFFER_LENGTH];
-  int retval, copydata=0;
-
-  // See struct hd_drive_cmd_hdr in hdreg.h
-  // buff[0]: ATA COMMAND CODE REGISTER
-  // buff[1]: ATA SECTOR NUMBER REGISTER == LBA LOW REGISTER
-  // buff[2]: ATA FEATURES REGISTER
-  // buff[3]: ATA SECTOR COUNT REGISTER
-  
-  // clear out buff.  Large enough for HDIO_DRIVE_CMD (4+512 bytes)
-  memset(buff, 0, STRANGE_BUFFER_LENGTH);
-
-  buff[0]=WIN_SMART;
-  switch (command){
-  case READ_VALUES:
-    buff[2]=SMART_READ_VALUES;
-    copydata=buff[3]=1;
-    break;
-  case READ_THRESHOLDS:
-    buff[2]=SMART_READ_THRESHOLDS;
-    copydata=buff[1]=buff[3]=1;
-    break;
-  case READ_LOG:
-    buff[2]=SMART_READ_LOG_SECTOR;
-    buff[1]=select;
-    copydata=buff[3]=1;
-    break;
-  case IDENTIFY:
-    buff[0]=WIN_IDENTIFY;
-    copydata=buff[3]=1;
-    break;
-  case PIDENTIFY:
-    buff[0]=WIN_PIDENTIFY;
-    copydata=buff[3]=1;
-    break;
-  case ENABLE:
-    buff[2]=SMART_ENABLE;
-    buff[1]=1;
-    break;
-  case DISABLE:
-    buff[2]=SMART_DISABLE;
-    buff[1]=1;
-    break;
-  case STATUS:
-    // this command only says if SMART is working.  It could be
-    // replaced with STATUS_CHECK below.
-    buff[2]=SMART_STATUS;
-    break;
-  case AUTO_OFFLINE:
-    buff[2]=SMART_AUTO_OFFLINE;
-    buff[3]=select;   // YET NOTE - THIS IS A NON-DATA COMMAND!!
-    break;
-  case AUTOSAVE:
-    buff[2]=SMART_AUTOSAVE;
-    buff[3]=select;   // YET NOTE - THIS IS A NON-DATA COMMAND!!
-    break;
-  case IMMEDIATE_OFFLINE:
-    buff[2]=SMART_IMMEDIATE_OFFLINE;
-    buff[1]=select;
-    break;
-    // the command uses HDIO_DRIVE_TASK and has different syntax than
-    // the other commands.
-  case STATUS_CHECK:
-    buff[1]=SMART_STATUS;
-    break;
-  default:
-    pout("Unrecognized command %d in linux_ata_command_interface()\n", command);
-    exit(1);
-    break;
-  }
-  
-  // There are two different types of ioctls().  The HDIO_DRIVE_TASK
-  // one is this:
-  if (command==STATUS_CHECK){
-    unsigned const char normal_lo=0x4f, normal_hi=0xc2;
-    unsigned const char failed_lo=0xf4, failed_hi=0x2c;
-    buff[4]=normal_lo;
-    buff[5]=normal_hi;
-    
-    // HDIO_DRIVE_TASK IOCTL
-#ifdef HDIO_DRIVE_TASK
-    if ((retval=ioctl(device, HDIO_DRIVE_TASK, buff)))
-#endif
-      return -1;
-    
-    // Cyl low and Cyl high unchanged means "Good SMART status"
-    if (buff[4]==normal_lo && buff[5]==normal_hi)
-      return 0;
-    
-    // These values mean "Bad SMART status"
-    if (buff[4]==failed_lo && buff[5]==failed_hi)
-      return 1;
-    
-    // We haven't gotten output that makes sense; print out some debugging info
-    syserror("Error SMART Status command failed");
-    pout("Please get assistance from %s\n",PROJECTHOME);
-    pout("Register values returned from SMART Status command are:\n");
-    pout("CMD=0x%02x\n",(int)buff[0]);
-    pout("FR =0x%02x\n",(int)buff[1]);
-    pout("NS =0x%02x\n",(int)buff[2]);
-    pout("SC =0x%02x\n",(int)buff[3]);
-    pout("CL =0x%02x\n",(int)buff[4]);
-    pout("CH =0x%02x\n",(int)buff[5]);
-    pout("SEL=0x%02x\n",(int)buff[6]);
-    return -1;   
-  }
-  
-  // We are now doing the HDIO_DRIVE_CMD type ioctl.
-  if ((retval=ioctl(device, HDIO_DRIVE_CMD, buff)))
-    return -1;
-  
-  if (copydata)
-    memcpy(data, buff+4, 512);
-  
-  return 0; 
-}
-
-
 static char *commandstrings[]={
-  [ENABLE]=           "SMART ENABLE",
-  [DISABLE]=          "SMART DISABLE",
-  [AUTOSAVE]=         "SMART AUTOMATIC ATTRIBUTE SAVE",
-  [IMMEDIATE_OFFLINE]="SMART IMMEDIATE OFFLINE",
-  [AUTO_OFFLINE]=     "SMART AUTO OFFLINE",
-  [STATUS]=           "SMART STATUS",
-  [STATUS_CHECK]=     "SMART STATUS CHECK",
-  [READ_VALUES]=      "SMART READ ATTRIBUTE VALUES",
-  [READ_THRESHOLDS]=  "SMART READ ATTRIBUTE THRESHOLDS",
-  [READ_LOG]=         "SMART READ LOG",
-  [IDENTIFY]=         "IDENTIFY DEVICE" ,
-  [PIDENTIFY]=        "IDENTIFY PACKET DEVICE" 
+  "SMART ENABLE",
+  "SMART DISABLE",
+  "SMART AUTOMATIC ATTRIBUTE SAVE",
+  "SMART IMMEDIATE OFFLINE",
+  "SMART AUTO OFFLINE",
+  "SMART STATUS",
+  "SMART STATUS CHECK",
+  "SMART READ ATTRIBUTE VALUES",
+  "SMART READ ATTRIBUTE THRESHOLDS",
+  "SMART READ LOG",
+  "IDENTIFY DEVICE",
+  "IDENTIFY PACKET DEVICE"
 };
 
 void prettyprint(unsigned char *stuff, char *name){
@@ -646,7 +502,9 @@ void prettyprint(unsigned char *stuff, char *name){
   pout("===== [%s] DATA END (512 Bytes) =====\n\n", name);
 }
 
-// This function provides the pretty-print reporting
+// This function provides the pretty-print reporting for SMART
+// commands: it implements the various -r "reporting" options for ATA
+// ioctls.
 int smartcommandhandler(int device, smart_command_set command, int select, char *data){
   int retval;
 
@@ -677,15 +535,30 @@ int smartcommandhandler(int device, smart_command_set command, int select, char
     return -1;
   }
   
+  // The reporting is cleaner, and we will find coding bugs faster, if
+  // the commands that failed clearly return empty (zeroed) data
+  // structures
+  if (getsdata)
+    memset(data, '\0', 512);
+
+  // In case the command produces an error, we'll want to know what it is:
+  errno=0;
+  
   // now execute the command
   if (con->escalade)
-    retval=linux_3ware_command_interface(device, con->escalade-1, command, select, data);
+    retval=escalade_command_interface(device, con->escalade-1, command, select, data);
   else
-    retval=linux_ata_command_interface(device, command, select, data);
+    retval=ata_command_interface(device, command, select, data);
   
   // If reporting is enabled, say what output was produced by the command
   if (con->reportataioctl){
-    pout("REPORT-IOCTL: DeviceFD=%d Command=%s returned %d\n", device, commandstrings[command], retval);
+    if (errno)
+      pout("REPORT-IOCTL: DeviceFD=%d Command=%s returned %d errno=%d [%s]\n", 
+	   device, commandstrings[command], retval, errno, strerror(errno));
+    else
+      pout("REPORT-IOCTL: DeviceFD=%d Command=%s returned %d\n",
+	   device, commandstrings[command], retval);
+    
     // if requested, pretty-print the output data structure
     if (con->reportataioctl>1 && getsdata)
       prettyprint((unsigned char *)data, commandstrings[command]);
@@ -707,53 +580,43 @@ unsigned char checksum(unsigned char *buffer){
   return sum;
 }
 
-// Reads current Device Identity info (512 bytes) into buf
-int ataReadHDIdentity (int device, struct hd_driveid *buf){
-  unsigned short driveidchecksum;
-  
-  if (smartcommandhandler(device, IDENTIFY, 0, (char *)buf)){
-    // See if device responds to packet command...
+// Reads current Device Identity info (512 bytes) into buf.  Returns 0
+// if all OK.  Returns -1 if no ATA Device identity can be
+// established.  Returns >0 if Device is ATA Packet Device (not SMART
+// capable).  The value of the integer helps identify the type of
+// Packet device, which is useful so that the user can connect the
+// formal device number with whatever object is inside their computer.
+int ataReadHDIdentity (int device, struct ata_identify_device *buf){
+  unsigned short *rawshort=(unsigned short *)buf;
+  unsigned char  *rawbyte =(unsigned char  *)buf;
+
+  // See if device responds either to IDENTIFY DEVICE or IDENTIFY
+  // PACKET DEVICE
+  if ((smartcommandhandler(device, IDENTIFY, 0, (char *)buf))){
     if (smartcommandhandler(device, PIDENTIFY, 0, (char *)buf)){
-      syserror("Error ATA GET HD Identity Failed");
       return -1; 
     }
   }
 
+  // if machine is big-endian, swap byte order as needed
   if (isbigendian()){
-    unsigned short *alias=(unsigned short*)buf;
     int i;
     
     // swap various capability words that are needed
-    swap2((char *)(alias+255)); 
+    swap2((char *)(rawshort+255)); 
     for (i=80; i<=87; i++)
-      swap2((char *)(alias+i));
+      swap2((char *)(rawshort+i));
   }
   
-#if 0
-  // The following ifdef is a HACK to distinguish different versions
-  // of the header file defining hd_driveid
-#ifdef CFA_REQ_EXT_ERROR_CODE
-  driveidchecksum=buf->integrity_word;
-#else
-  // Note -- the declaration that appears in
-  // /usr/include/linux/hdreg.h: short words160_255[95], is WRONG.
-  // It should say: short words160_255[96]. I have written to Andre
-  // Hedrick about this on Oct 17 2002.  Please remove this comment
-  // once the fix has made it into the stock kernel tree.
-  driveidchecksum=buf->words160_255[95];
-#endif
-#else
-  // This way is ugly and you may feel ill -- but it always works...
-  {
-    unsigned short *rawstructure=
-      (unsigned short *)buf;
-    driveidchecksum=rawstructure[255];
-  }
-#endif
-  
-  if ((driveidchecksum & 0x00ff) == 0x00a5 && checksum((unsigned char *)buf))
+  // If there is a checksum there, validate it
+  if ((rawshort[255] & 0x00ff) == 0x00a5 && checksum(rawbyte))
     checksumwarning("Drive Identity Structure");
- 
+  
+  // If this is a PACKET DEVICE, return device type
+  if (rawbyte[1] & 0x80)
+    return 1+(rawbyte[1] & 0x1f);
+  
+  // Not a PACKET DEVICE
   return 0;
 }
 
@@ -761,7 +624,7 @@ int ataReadHDIdentity (int device, struct hd_driveid *buf){
 // describing which revision.  Note that Revision 0 of ATA-3 does NOT
 // support SMART.  For this one case we return -3 rather than +3 as
 // the version number.  See notes above.
-int ataVersionInfo (const char** description, struct hd_driveid *drive, unsigned short *minor){
+int ataVersionInfo (const char** description, struct ata_identify_device *drive, unsigned short *minor){
   unsigned short major;
   int i;
 
@@ -783,13 +646,8 @@ int ataVersionInfo (const char** description, struct hd_driveid *drive, unsigned
   }
 
   // get major and minor ATA revision numbers
-#ifdef __NEW_HD_DRIVE_ID
   major=drive->major_rev_num;
   *minor=drive->minor_rev_num;
-#else
-  major=drive->word80;
-  *minor=drive->word81;
-#endif
   
   // First check if device has ANY ATA version information in it
   if (major==NOVAL_0 || major==NOVAL_1) {
@@ -820,39 +678,28 @@ int ataVersionInfo (const char** description, struct hd_driveid *drive, unsigned
     return i;;
 }
 
-// returns 1 if SMART supported, 0 if not supported or can't tell
-int ataSmartSupport(struct hd_driveid *drive){
-  unsigned short word82,word83;
-
-  // get correct bits of IDENTIFY DEVICE structure
-#ifdef __NEW_HD_DRIVE_ID
-  word82=drive->command_set_1;
-  word83=drive->command_set_2;
-#else
-  word82=drive->command_sets;
-  word83=drive->word83;
-#endif
-
-  // Note this does not work for ATA3 < Revision 6, when word82 and word83 were added
-  // we should check for ATA3 Rev 0 in minor identity code...  
-  return (word83 & 0x0001<<14) && !(word83 & 0x0001<<15) && (word82 & 0x0001);
+// returns 1 if SMART supported, 0 if SMART unsupported, -1 if can't tell
+int ataSmartSupport(struct ata_identify_device *drive){
+  unsigned short word82=drive->command_set_1;
+  unsigned short word83=drive->command_set_2;
+  
+  // check if words 82/83 contain valid info
+  if ((word83>>14) == 0x01)
+    // return value of SMART support bit 
+    return word82 & 0x0001;
+  
+  // since we can're rely on word 82, we don't know if SMART supported
+  return -1;
 }
 
 // returns 1 if SMART enabled, 0 if SMART disabled, -1 if can't tell
-int ataIsSmartEnabled(struct hd_driveid *drive){
-    unsigned short word85,word87;
-
-  // Get correct bits of IDENTIFY DRIVE structure
-#ifdef __NEW_HD_DRIVE_ID
-  word85=drive->cfs_enable_1;
-  word87=drive->csf_default;
-#else
-  word85=drive->word85;
-  word87=drive->word87;
-#endif
+int ataIsSmartEnabled(struct ata_identify_device *drive){
+  unsigned short word85=drive->cfs_enable_1;
+  unsigned short word87=drive->csf_default;
   
-  if ((word87 & 0x0001<<14) && !(word87 & 0x0001<<15))
-    // word85 contains valid information, so
+  // check if words 85/86/87 contain valid info
+  if ((word87>>14) == 0x01)
+    // return value of SMART enabled bit
     return word85 & 0x0001;
   
   // Since we can't rely word85, we don't know if SMART is enabled.
@@ -905,8 +752,8 @@ void fixsamsungselftestlog(struct ata_smart_selftestlog *data){
 }
 
 // Reads the Self Test Log (log #6)
-int ataReadSelfTestLog (int device, struct ata_smart_selftestlog *data){	
-  
+int ataReadSelfTestLog (int device, struct ata_smart_selftestlog *data){
+
   // get data from device
   if (smartcommandhandler(device, READ_LOG, 0x06, (char *)data)){
     syserror("Error SMART Error Self-Test Log Read failed");
@@ -952,6 +799,7 @@ int ataReadLogDirectory (int device, struct ata_smart_log_directory *data){
   return 0;
 }
 
+
 // Reads the selective self-test log (log #9)
 int ataReadSelectiveSelfTestLog(int device, struct ata_selective_self_test_log *data){	
   
@@ -976,14 +824,71 @@ int ataReadSelectiveSelfTestLog(int device, struct ata_selective_self_test_log *
   return 0;
 }
 
+#if DEVELOP_SELECTIVE_SELF_TEST
+// Writes the selective self-test log (log #9)
+int ataWriteSelectiveSelfTestLog(int device){	
+  int i;
+  struct ata_selective_self_test_log sstlog, *data=&sstlog;
+  unsigned char cksum=0;
+  unsigned char *ptr=(unsigned char *)data;
+
+  // Read log
+  ataReadSelectiveSelfTestLog(device, data);
+
+  // Clear spans
+  for (i=0; i<5; i++)
+    memset(data->span+i, 0, sizeof(struct test_span));
+
+  // Set spans for testing 
+  for (i=0; i<con->smartselectivenumspans; i++){
+    data->span[i].start = con->smartselectivespan[i][0];
+    data->span[i].end   = con->smartselectivespan[i][1];
+  }
+
+  // Do NOT perform off-line scan after selective 
+  data->flags=0;
+  data->undefined=data->checksum=0;
+
+  // Put in correct checksum
+  for (i=0; i<512; i++)
+    cksum+=ptr[i];
+  cksum=~cksum;
+  cksum+=1;
+  data->checksum=cksum;
+
+    // swap endian order if needed
+  if (isbigendian()){
+    int i;
+    swap2((char *)&(data->logversion));
+    for (i=0;i<5;i++){
+      swap8((char *)&(data->span[i].start));
+      swap8((char *)&(data->span[i].end));
+    }
+    swap8((char *)&(data->currentlba));
+    swap2((char *)&(data->currentspan));
+    swap2((char *)&(data->flags));
+    swap2((char *)&(data->pendingtime));
+  }
+
+  // send data to device -- ioctl NOT implemented yet!
+  if (smartcommandhandler(device, WRITE_LOG, 0x09, (char *)data)){
+    return -1;
+  }
+
+  return 0;
+}
+#endif
+
 // This corrects some quantities that are byte reversed in the SMART
-// ATA ERROR LOG
+// ATA ERROR LOG.
 void fixsamsungerrorlog(struct ata_smart_errorlog *data){
   int i,j;
   
+  // FIXED IN SAMSUNG -25 FIRMWARE???
   // Device error count in bytes 452-3
   swap2((char *)&(data->ata_error_count));
   
+  // FIXED IN SAMSUNG -22a FIRMWARE
   // step through 5 error log data structures
   for (i=0; i<5; i++){
     // step through 5 command data structures
@@ -996,6 +901,13 @@ void fixsamsungerrorlog(struct ata_smart_errorlog *data){
   return;
 }
 
+// NEEDED ONLY FOR SAMSUNG -22 (some) -23 AND -24?? FIRMWARE
+void fixsamsungerrorlog2(struct ata_smart_errorlog *data){
+  // Device error count in bytes 452-3
+  swap2((char *)&(data->ata_error_count));
+  return;
+}
+
 // Reads the Summary SMART Error Log (log #1). The Comprehensive SMART
 // Error Log is #2, and the Extended Comprehensive SMART Error log is
 // #3
@@ -1015,6 +927,8 @@ int ataReadErrorLog (int device, struct ata_smart_errorlog *data){
   // Error log entries
   if (con->fixfirmwarebug == FIX_SAMSUNG)
     fixsamsungerrorlog(data);
+  else if (con->fixfirmwarebug == FIX_SAMSUNG2)
+    fixsamsungerrorlog2(data);
 
   // Correct endian order if necessary
   if (isbigendian()){
@@ -1090,16 +1004,10 @@ int ataDisableAutoSave(int device){
   return 0;
 }
 
-// Note that in the ATA-5 standard the Enable/Disable AutoOffline
-// command is marked "OBSOLETE".  Curiously, I could not find it
-// documented in ANY of the ATA specifications.  In other words, it's
-// been obsolete forever. However some vendors (eg, IBM) seem to be
-// using this command anyway.  For example see the IBM Travelstar
-// 40GNX hard disk drive specifications page 164 Revision 1.1 22 Apr
-// 2002.  This gives a detailed description of the command, although
-// the drive claims to comply with the ATA/ATAPI-5 Revision 3
-// standard!  The latter document makes no mention of this command at
-// all, other than to say that it is "obsolete".
+// In *ALL* ATA standards the Enable/Disable AutoOffline command is
+// marked "OBSOLETE". It is defined in SFF-8035i Revision 2, and most
+// vendors still support it for backwards compatibility. IBM documents
+// it for some drives.
 int ataEnableAutoOffline (int device ){	
   
   /* timer hard coded to 4 hours */  
@@ -1128,7 +1036,6 @@ int ataDoesSmartWork(int device){
   return !smartcommandhandler(device, STATUS, 0, NULL);
 }
 
-#ifdef HDIO_DRIVE_TASK
 // This function uses a different interface (DRIVE_TASK) than the
 // other commands in this file.
 int ataSmartStatus2(int device){
@@ -1142,31 +1049,20 @@ int ataSmartStatus2(int device){
   
   return returnval;
 }
-#else
-// Just a hack so that the code compiles on 
-// 2.2 kernels without HDIO_DRIVE TASK support.  
-// Should be fixed by putting in a call to code 
-// that compares smart data to thresholds.
-int ataSmartStatus2(int device){
-  pout("This code was compiled on a machine whose kernel header\n"
-       "files do not support the HDIO_DRIVE_TASK ioctl().\n"
-       "Compile on a linux 2.2 kernel box with HDIO_DRIVE_TASK\n"
-       "support enabled, or on a 2.4 kernel box, please.\n");
-  return ataSmartStatus(device);
-}
-#endif
-
 
 // This is the way to execute ALL tests: offline, short self-test,
 // extended self test, with and without captive mode, etc.
 int ataSmartTest(int device, int testtype){	
   char cmdmsg[128],*type,*captive;
   int errornum;
-  int cap;
+  int cap,select=0;
 
   // Boolean, if set, says test is captive
   cap=testtype & CAPTIVE_MASK;
   
+  // Boolean, if set, then test is selective
+  select=(testtype==SELECTIVE_SELF_TEST || testtype==SELECTIVE_CAPTIVE_SELF_TEST);
+
   // Set up strings that describe the type of test
   if (cap)
     captive="captive";
@@ -1181,11 +1077,20 @@ int ataSmartTest(int device, int testtype){
     type="Extended self-test";
   else if (testtype==CONVEYANCE_SELF_TEST || testtype==CONVEYANCE_CAPTIVE_SELF_TEST)
     type="Conveyance self-test";
-  else if (testtype==SELECTIVE_SELF_TEST || testtype==SELECTIVE_CAPTIVE_SELF_TEST)
+#if DEVELOP_SELECTIVE_SELF_TEST
+  else if (select){
+    int i;
     type="Selective self-test";
+    pout("Using test spans:\n");
+    for (i = 0; i < con->smartselectivenumspans; i++)
+      pout("%lld - %lld\n",
+	   con->smartselectivespan[i][0],
+	   con->smartselectivespan[i][1]);
+  }
+#endif
   else
     type="[Unrecognized] self-test";
-    
+  
   //  Print ouf message that we are sending the command to test
   if (testtype==ABORT_SELF_TEST)
     sprintf(cmdmsg,"Abort SMART off-line mode self-test routine");
@@ -1193,6 +1098,17 @@ int ataSmartTest(int device, int testtype){
     sprintf(cmdmsg,"Execute SMART %s routine immediately in %s mode",type,captive);
   pout("Sending command: \"%s\".\n",cmdmsg);
 
+#if DEVELOP_SELECTIVE_SELF_TEST
+  // If doing a selective self-test, issue WRITE_LOG command
+  if (select && ataWriteSelectiveSelfTestLog(device)){
+    char errormsg[256];
+    sprintf(errormsg,"Command \"%s\" failed to WRITE Selective Self-Test Log ",cmdmsg); 
+    syserror(errormsg);
+    pout("\n");
+    return -1;
+  }
+#endif
+
   // Now send the command to test
   errornum=smartcommandhandler(device, IMMEDIATE_OFFLINE, testtype, NULL);
 
@@ -1233,20 +1149,17 @@ int TestTime(struct ata_smart_values *data,int testtype){
 
 // This function tells you both about the ATA error log and the
 // self-test error log capability.  The bit is poorly documented in
-// the ATA/ATAPI standard.
+// the ATA/ATAPI standard.  SMART error logging is also indicated in
+// bit 0 of DEVICE IDENTIFY word 87 (if top two bits of word 87 match
+// pattern 01).  However this was only introduced in ATA-6 (but error
+// logging was in ATA-5).
 int isSmartErrorLogCapable ( struct ata_smart_values *data){
    return data->errorlog_capability & 0x01;
 }
-int isSupportExecuteOfflineImmediate(struct ata_smart_values *data){
-   return data->offline_data_collection_capability & 0x01;
-}
 
-int isGeneralPurposeLoggingCapable(struct hd_driveid *identity){
-  unsigned short *rawwords=(unsigned short *)identity;
-  unsigned short word84, word87;
-
-  word84=rawwords[84];
-  word87=rawwords[87];
+int isGeneralPurposeLoggingCapable(struct ata_identify_device *identity){
+  unsigned short word84=identity->command_set_extension;
+  unsigned short word87=identity->csf_default;
 
   // If bit 14 of word 84 is set to one and bit 15 of word 84 is
   // cleared to zero, the contents of word 84 contains valid support
@@ -1270,6 +1183,13 @@ int isGeneralPurposeLoggingCapable(struct hd_driveid *identity){
 }
 
 
+// SMART self-test capability is also indicated in bit 1 of DEVICE
+// IDENTIFY word 87 (if top two bits of word 87 match pattern 01).
+// However this was only introduced in ATA-6 (but self-test log was in
+// ATA-5).
+int isSupportExecuteOfflineImmediate(struct ata_smart_values *data){
+   return data->offline_data_collection_capability & 0x01;
+}
 // Note in the ATA-5 standard, the following bit is listed as "Vendor
 // Specific".  So it may not be reliable. The only use of this that I
 // have found is in IBM drives, where it is well-documented.  See for
@@ -1663,23 +1583,39 @@ void ataPrintSmartAttribName(char *out, unsigned char id, unsigned char *definit
   case 202:
     // Fujitsu
     name="TA_Increase_Count";
+    // Maxtor: Data Address Mark Errors
     break;
   case 203:
     // Fujitsu
     name="Run_Out_Cancel";
+    // Maxtor: ECC Errors
     break;
   case 204:
     // Fujitsu
     name="Shock_Count_Write_Opern";
+    // Maxtor: Soft ECC Correction
     break;
   case 205:
     // Fujitsu
     name="Shock_Rate_Write_Opern";
+    // Maxtor: Thermal Aspirates
     break;
   case 206:
     // Fujitsu
     name="Flying_Height_Measurement";
     break;
+  case 207:
+    // Maxtor
+    name="Spin_High_Current";
+    break;
+  case 208:
+    // Maxtor
+    name="Spin_Buzz";
+    break;
+  case 209:
+    // Maxtor
+    name="Offline_Seek_Performnce";
+    break;
   case 220:
     switch (val) {
     case 1:
@@ -1731,6 +1667,6 @@ void ataPrintSmartAttribName(char *out, unsigned char id, unsigned char *definit
     name="Unknown_Attribute";
     break;
   }
-  sprintf(out,"%3hhu %s",id,name);
+  sprintf(out,"%3hu %s",(short int)id,name);
   return;
 }
