@@ -22,15 +22,12 @@
  *
  */
 
-#ifndef _ATACMDS_H_
-#define _ATACMDS_H_
+#ifndef ATACMDS_H_
+#define ATACMDS_H_
 
-#ifndef ATACMDS_H_CVSID
-#define ATACMDS_H_CVSID "$Id: atacmds.h,v 1.49 2003/08/13 12:33:23 ballen4705 Exp $\n"
-#endif
+#define ATACMDS_H_CVSID "$Id: atacmds.h,v 1.55 2003/10/15 14:06:02 ballen4705 Exp $\n"
 
 #include <sys/ioctl.h>
-#include <linux/hdreg.h>
 #include <sys/fcntl.h>
 #include <sys/types.h>
 
@@ -50,65 +47,35 @@ typedef enum {
   READ_LOG,
   IDENTIFY,
   PIDENTIFY
+#if DEVELOP_SELECTIVE_SELF_TEST
+  // writes 512 bytes of data:
+  , WRITE_LOG
+#endif
 } smart_command_set;
 
-/* These defines SHOULD BE in the kernel
-   if not we define them */
-							  
-#ifndef WIN_SMART
-#define WIN_SMART		0xb0	
-#endif
+// ATA Specification Command Register Values (Commands)
+#define WIN_IDENTIFY            0xec						  
+#define WIN_PIDENTIFY		0xa1
+#define WIN_SMART		0xb0
 
-#ifndef SMART_READ_VALUES
+// ATA Specification Feature Register Values (SMART Subcommands).
+// Note that some are obsolete as of ATA-7.
 #define SMART_READ_VALUES	0xd0
-#endif
-
-#ifndef SMART_READ_THRESHOLDS	
 #define SMART_READ_THRESHOLDS	0xd1
-#endif
-
-#ifndef SMART_AUTOSAVE 
 #define SMART_AUTOSAVE		0xd2
-#endif
-
-#ifndef SMART_SAVE
 #define SMART_SAVE		0xd3
-#endif
-
-#ifndef SMART_IMMEDIATE_OFFLINE	
 #define SMART_IMMEDIATE_OFFLINE	0xd4
-#endif
-
-#ifndef SMART_READ_LOG_SECTOR
-#define SMART_READ_LOG_SECTOR 0xd5
-#endif
-
-#ifndef SMART_WRITE_LOG_SECTOR
-#define SMART_WRITE_LOG_SECTOR 0xd6
-#endif
-
-// The following is obsolete -- don't use it!
-#ifndef SMART_WRITE_THRESHOLDS
+#define SMART_READ_LOG_SECTOR   0xd5
+#define SMART_WRITE_LOG_SECTOR  0xd6
 #define SMART_WRITE_THRESHOLDS  0xd7
-#endif
-
-#ifndef SMART_ENABLE
 #define SMART_ENABLE		0xd8
-#endif
-
-#ifndef SMART_DISABLE
 #define SMART_DISABLE		0xd9
-#endif
-
-#ifndef SMART_STATUS
 #define SMART_STATUS		0xda
-#endif
-
-// The following is also marked obsolete in ATA-5
-#ifndef SMART_AUTO_OFFLINE
+// SFF 8035i Revision 2 Specification Feature Register Value (SMART
+// Subcommand)
 #define SMART_AUTO_OFFLINE	0xdb
-#endif
 
+// Sector Number values for SMART_IMMEDIATE_OFFLINE Subcommand
 #define OFFLINE_FULL_SCAN		0
 #define SHORT_SELF_TEST			1
 #define EXTEND_SELF_TEST		2
@@ -121,28 +88,42 @@ typedef enum {
 #define SELECTIVE_CAPTIVE_SELF_TEST     132
 #define CAPTIVE_MASK                    (0x01<<7)
 
+// Maximum allowed number of SMART Attributes
 #define NUMBER_ATA_SMART_ATTRIBUTES 	30
 
-#define ATA_SMART_SEC_SIZE		512
-
-#ifndef HDIO_DRIVE_CMD_HDR_SIZE
-#define HDIO_DRIVE_CMD_HDR_SIZE 	4
-#endif
-
-#ifndef HDIO_DRIVE_TASK_HDR_SIZE
-#define HDIO_DRIVE_TASK_HDR_SIZE	7
-#endif
+// Needed parts of the ATA DRIVE IDENTIFY Structure. Those labeled
+// word* are NOT used.
+struct ata_identify_device {
+  unsigned short words000_009[10];
+  unsigned char	 serial_no[20];
+  unsigned short words020_022[3];
+  unsigned char	 fw_rev[8];
+  unsigned char	 model[40];
+  unsigned short words047_079[33];
+  unsigned short major_rev_num;
+  unsigned short minor_rev_num;
+  unsigned short command_set_1;
+  unsigned short command_set_2;
+  unsigned short command_set_extension;
+  unsigned short cfs_enable_1;
+  unsigned short word086;
+  unsigned short csf_default;
+  unsigned short words088_255[168];
+};
 
 /* ata_smart_attribute is the vendor specific in SFF-8035 spec */ 
+#pragma pack(1)
 struct ata_smart_attribute {
   unsigned char id;
   // meaning of flag bits: see MACROS just below
+  // WARNING: MISALIGNED!
   unsigned short flags; 
   unsigned char current;
   unsigned char worst;
   unsigned char raw[6];
   unsigned char reserv;
-} __attribute__ ((packed));
+};
+#pragma pack()
 
 // MACROS to interpret the flags bits in the previous structure.
 // These have not been implemented using bitflags and a union, to make
@@ -188,28 +169,9 @@ struct ata_smart_attribute {
 
 // Last ten bits are reserved for future use
 
-
-// What follows is the ATA/ATAPI-4 Rev 13 data structure equivalent:
-// Table 23 DEVICE SMART DATA STRUCTURE
-#if 0
-struct ata_smart_values {
-  unsigned short int revnumber;
-  struct ata_smart_attribute vendor_attributes [NUMBER_ATA_SMART_ATTRIBUTES];
-  unsigned char offline_data_collection_status;
-  unsigned char vendor_specific_363;  //IBM # segments for offline collection
-  unsigned short int total_time_to_complete_off_line; // IBM different
-  unsigned char vendor_specific_366; // IBM curent segment pointer
-  unsigned char offline_data_collection_capability;
-  unsigned short int smart_capability;
-  unsigned char reserved_370_385;;
-  unsigned char vendor_specific_386-510;  // IBM: self-test failure checkpoint
-  unsigned char chksum;
-} __attribute__ ((packed));
-#endif
-
-
 /* ata_smart_values is format of the read drive Attribute command */
 /* see Table 34 of T13/1321D Rev 1 spec (Device SMART data structure) for *some* info */
+#pragma pack(1)
 struct ata_smart_values {
   unsigned short int revnumber;
   struct ata_smart_attribute vendor_attributes [NUMBER_ATA_SMART_ATTRIBUTES];
@@ -227,29 +189,32 @@ struct ata_smart_values {
   unsigned char reserved_375_385[11];
   unsigned char vendor_specific_386_510[125];
   unsigned char chksum;
-} __attribute__ ((packed));
-
-/* Smart Threshold data structures */
+}; 
+#pragma pack()
 
 /* Vendor attribute of SMART Threshold (compare to ata_smart_attribute above) */
+#pragma pack(1)
 struct ata_smart_threshold_entry {
   unsigned char id;
   unsigned char threshold;
   unsigned char reserved[10];
-} __attribute__ ((packed));
-
+};
+#pragma pack()
 
 /* Format of Read SMART THreshold Command */
 /* Compare to ata_smart_values above */
+#pragma pack(1)
 struct ata_smart_thresholds {
   unsigned short int revnumber;
   struct ata_smart_threshold_entry thres_entries[NUMBER_ATA_SMART_ATTRIBUTES];
   unsigned char reserved[149];
   unsigned char chksum;
-} __attribute__ ((packed));
+};
+#pragma pack()
 
 
 // Table 42 of T13/1321D Rev 1 spec (Error Data Structure)
+#pragma pack(1)
 struct ata_smart_errorlog_error_struct {
   unsigned char reserved;
   unsigned char error_register;
@@ -262,9 +227,12 @@ struct ata_smart_errorlog_error_struct {
   unsigned char extended_error[19];
   unsigned char state;
   unsigned short timestamp;
-} __attribute__ ((packed));
+};
+#pragma pack()
+
 
 // Table 41 of T13/1321D Rev 1 spec (Command Data Structure)
+#pragma pack(1)
 struct ata_smart_errorlog_command_struct {
   unsigned char devicecontrolreg;
   unsigned char featuresreg;
@@ -275,15 +243,19 @@ struct ata_smart_errorlog_command_struct {
   unsigned char drive_head;
   unsigned char commandreg;
   unsigned int timestamp;
-} __attribute__ ((packed));
+};
+#pragma pack()
 
 // Table 40 of T13/1321D Rev 1 spec (Error log data structure)
+#pragma pack(1)
 struct ata_smart_errorlog_struct {
   struct ata_smart_errorlog_command_struct commands[5];
   struct ata_smart_errorlog_error_struct error_struct;
-}  __attribute__ ((packed));
+};
+#pragma pack()
 
 // Table 39 of T13/1321D Rev 1 spec (SMART error log sector)
+#pragma pack(1)
 struct ata_smart_errorlog {
   unsigned char revnumber;
   unsigned char error_log_pointer;
@@ -291,9 +263,11 @@ struct ata_smart_errorlog {
   unsigned short int ata_error_count;
   unsigned char reserved[57];
   unsigned char checksum;
-} __attribute__ ((packed));
+};
+#pragma pack()
 
 // Table 45 of T13/1321D Rev 1 spec (Self-test log descriptor entry)
+#pragma pack(1)
 struct ata_smart_selftestlog_struct {
   unsigned char selftestnumber; // Sector number register
   unsigned char selfteststatus;
@@ -301,9 +275,11 @@ struct ata_smart_selftestlog_struct {
   unsigned char selftestfailurecheckpoint;
   unsigned int lbafirstfailure;
   unsigned char vendorspecific[15];
-} __attribute__ ((packed));
+};
+#pragma pack()
 
 // Table 44 of T13/1321D Rev 1 spec (Self-test log data structure)
+#pragma pack(1)
 struct ata_smart_selftestlog {
   unsigned short int revnumber;
   struct ata_smart_selftestlog_struct selftest_struct[21];
@@ -311,29 +287,34 @@ struct ata_smart_selftestlog {
   unsigned char mostrecenttest;
   unsigned char reserved[2];
   unsigned char chksum;
-} __attribute__ ((packed));
-
-
+};
+#pragma pack()
 
 // SMART LOG DIRECTORY Table 52 of T13/1532D Vol 1 Rev 1a
+#pragma pack(1)
 struct ata_smart_log_entry {
   unsigned char numsectors;
   unsigned char reserved;
-} __attribute__ ((packed));
+};
+#pragma pack()
 
+#pragma pack(1)
 struct ata_smart_log_directory {
   unsigned short int logversion;
   struct ata_smart_log_entry entry[255];
-} __attribute__ ((packed));
-
+};
+#pragma pack()
 
 // SMART SELECTIVE SELF-TEST LOG Table 61 of T13/1532D Volume 1
 // Revision 3
+#pragma pack(1)
 struct test_span {
   unsigned long long start;
   unsigned long long end;
-} __attribute__ ((packed));
+};
+#pragma pack()
 
+#pragma pack(1)
 struct ata_selective_self_test_log {
   unsigned short     logversion;
   struct test_span   span[5];
@@ -346,10 +327,11 @@ struct ata_selective_self_test_log {
   unsigned short     pendingtime;
   unsigned char      undefined;  // Error in specs!
   unsigned char      checksum;
-} __attribute__ ((packed));
+};
+#pragma pack()
 
 /* Read S.M.A.R.T information from drive */
-int ataReadHDIdentity(int device, struct hd_driveid *buf);
+int ataReadHDIdentity(int device, struct ata_identify_device *buf);
 int ataReadSmartValues(int device,struct ata_smart_values *);
 int ataReadSmartThresholds(int device, struct ata_smart_thresholds *);
 int ataReadErrorLog(int device, struct ata_smart_errorlog *);
@@ -378,19 +360,19 @@ int ataSmartSelfTestAbort (int device);
 
 // Returns the latest compatibility of ATA/ATAPI Version the device
 // supports. Returns -1 if Version command is not supported
-int ataVersionInfo (const char **description, struct hd_driveid *drive, unsigned short *minor);
+int ataVersionInfo (const char **description, struct ata_identify_device *drive, unsigned short *minor);
 
 // If SMART supported, this is guaranteed to return 1 if SMART is enabled, else 0.
 int ataDoesSmartWork(int device);
 
 // returns 1 if SMART supported, 0 if not supported or can't tell
-int ataSmartSupport ( struct hd_driveid *drive);
+int ataSmartSupport ( struct ata_identify_device *drive);
 
 // Return values:
 //  1: SMART enabled
 //  0: SMART disabled
 // -1: can't tell if SMART is enabled -- try issuing ataDoesSmartWork command to see
-int ataIsSmartEnabled(struct hd_driveid *drive);
+int ataIsSmartEnabled(struct ata_identify_device *drive);
 
 /* Check SMART for Threshold failure */
 // onlyfailed=0 : are or were any age or prefailure attributes <= threshold
@@ -409,7 +391,7 @@ int isExtendedSelfTestTime ( struct ata_smart_values *data);
 
 int isSmartErrorLogCapable ( struct ata_smart_values *data);
 
-int isGeneralPurposeLoggingCapable(struct hd_driveid *identity);
+int isGeneralPurposeLoggingCapable(struct ata_identify_device *identity);
 
 int isSupportExecuteOfflineImmediate ( struct ata_smart_values *data);
 
@@ -471,5 +453,10 @@ int parse_attribute_def(char *pair, unsigned char **defs);
 char *create_vendor_attribute_arg_list(void);
 
 
+// These are two of the functions that are defined in os_*.c and need
+// to be ported to get smartmontools onto another OS.
+int ata_command_interface(int device, smart_command_set command, int select, char *data);
+int escalade_command_interface(int fd, int disknum, smart_command_set command, int select, char *data);
 
-#endif /* _ATACMDS_H_ */
+
+#endif /* ATACMDS_H_ */
