@@ -33,13 +33,17 @@
 #include "extern.h"
 #include "utility.h"
 #include "knowndrives.h"
+#include "config.h"
 
-const char *ataprint_c_cvsid="$Id: ataprint.c,v 1.97 2003/08/13 12:33:23 ballen4705 Exp $"
-ATACMDNAMES_H_CVSID ATACMDS_H_CVSID ATAPRINT_H_CVSID EXTERN_H_CVSID KNOWNDRIVES_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
+const char *ataprint_c_cvsid="$Id: ataprint.c,v 1.115 2003/11/26 20:42:49 pjwilliams Exp $"
+ATACMDNAMES_H_CVSID ATACMDS_H_CVSID ATAPRINT_H_CVSID CONFIG_H_CVSID EXTERN_H_CVSID KNOWNDRIVES_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
 
 // for passing global control variables
 extern smartmonctrl *con;
 
+// to hold onto exit code for atexit routine
+extern int exitstatus;
+
 // Copies n bytes (or n-1 if n is odd) from in to out, but swaps adjacents
 // bytes.
 void swapbytes(char *out, const char *in, size_t n)
@@ -60,7 +64,7 @@ void trim(char *out, const char *in)
   // Find the first non-space character (maybe none).
   first = -1;
   for (i = 0; in[i]; i++)
-    if (!isspace(in[i])) {
+    if (!isspace((int)in[i])) {
       first = i;
       break;
     }
@@ -72,7 +76,7 @@ void trim(char *out, const char *in)
   }
 
   // Find the last non-space character.
-  for (i = strlen(in)-1; i >= first && isspace(in[i]); i--)
+  for (i = strlen(in)-1; i >= first && isspace((int)in[i]); i--)
     ;
   last = i;
 
@@ -80,7 +84,7 @@ void trim(char *out, const char *in)
   out[last-first+1] = '\0';
 }
 
-// Convenience function for formatting strings from hd_driveid.
+// Convenience function for formatting strings from ata_identify_device
 void formatdriveidstring(char *out, const char *in, int n)
 {
   char tmp[65];
@@ -102,7 +106,139 @@ void printswap(char *output, char *in, unsigned int n){
     pout("[No Information Found]\n");
 }
 
-void ataPrintDriveInfo (struct hd_driveid *drive){
+/* For the given Command Register (CR) and Features Register (FR), attempts
+ * to construct a string that describes the contents of the Status
+ * Register (ST) and Error Register (ER).  The string is dynamically allocated
+ * memory and the return value is a pointer to this string.  It is up to the
+ * caller to free this memory.  If there is insufficient memory or if the
+ * meanings of the flags of the error register are not known for the given
+ * command then it returns NULL.
+ *
+ * The meanings of the flags of the error register for all commands are
+ * described in the ATA spec and could all be supported here in theory.
+ * Currently, only a few commands are supported (those that have been seen
+ * to produce errors).  If many more are to be added then this function
+ * should probably be redesigned.
+ */
+char *construct_st_er_desc(unsigned char CR, unsigned char FR,
+                                 unsigned char ST, unsigned char ER)
+{
+  char *s;
+  char *error_flag[8];
+  int i;
+  /* If for any command the Device Fault flag of the status register is
+   * not used then used_device_fault should be set to 0 (in the CR switch
+   * below)
+   */
+  int uses_device_fault = 1;
+
+  /* A value of NULL means that the error flag isn't used */
+  for (i = 0; i < 8; i++)
+    error_flag[i] = NULL;
+
+  switch (CR) {
+  case 0x20:  /* READ SECTOR(S) */
+    error_flag[6] = "UNC";
+    error_flag[5] = "MC";
+    error_flag[4] = "IDNF";
+    error_flag[3] = "MCR";
+    error_flag[2] = "ABRT";
+    error_flag[1] = "NM";
+    error_flag[0] = "obs";
+    break;
+  case 0x25:  /* READ DMA EXT */
+  case 0xC8:  /* READ DMA */
+    error_flag[7] = "ICRC";
+    error_flag[6] = "UNC";
+    error_flag[5] = "MC";
+    error_flag[4] = "IDNF";
+    error_flag[3] = "MCR";
+    error_flag[2] = "ABRT";
+    error_flag[1] = "NM";
+    error_flag[0] = "obs";
+    break;
+  case 0x30:  /* WRITE SECTOR(S) */
+  case 0xC5:  /* WRITE MULTIPLE */
+    error_flag[6] = "WP";
+    error_flag[5] = "MC";
+    error_flag[4] = "IDNF";
+    error_flag[3] = "MCR";
+    error_flag[2] = "ABRT";
+    error_flag[1] = "NM";
+    break;
+  case 0xA1:  /* IDENTIFY PACKET DEVICE */
+  case 0xEF:  /* SET FEATURES */
+  case 0x00:  /* NOP */
+    error_flag[2] = "ABRT";
+    break;
+  case 0xB0:  /* SMART */
+    switch(FR) {
+    case 0xD6:  /* SMART WRITE LOG */
+      error_flag[4] = "IDNF";
+      error_flag[2] = "ABRT";
+      break;
+    case 0xD9:  /* SMART DISABLE OPERATIONS */
+    case 0xDA:  /* SMART RETURN STATUS */
+      error_flag[2] = "ABRT";
+      break;
+    default:
+      return NULL;
+      break;
+    }
+    break;
+  case 0xB1:  /* DEVICE CONFIGURATION */
+    switch (FR) {
+    case 0xC0:  /* DEVICE CONFIGURATION RESTORE */
+      error_flag[2] = "ABRT";
+      break;
+    default:
+      return NULL;
+      break;
+    }
+    break;
+  case 0xCA:  /* WRITE DMA */
+    error_flag[7] = "ICRC";
+    error_flag[6] = "WP";
+    error_flag[5] = "MC";
+    error_flag[4] = "IDNF";
+    error_flag[3] = "MCR";
+    error_flag[2] = "ABRT";
+    error_flag[1] = "NM";
+    error_flag[0] = "obs";
+    break;
+  default:
+    return NULL;
+  }
+
+  /* 100 bytes -- that'll be plenty (OK, this is lazy!) */
+  if (!(s = (char *)malloc(100)))
+    return s;
+
+  s[0] = '\0';
+
+  /* We ignore any status flags other than Device Fault and Error */
+
+  if (uses_device_fault && (ST & (1 << 5))) {
+    strcat(s, "Device Fault");
+    if (ST & 1)  // Error flag
+      strcat(s, "; ");
+  }
+  if (ST & 1) {  // Error flag
+    int count = 0;
+
+    strcat(s, "Error: ");
+    for (i = 7; i >= 0; i--)
+      if ((ER & (1 << i)) && (error_flag[i])) {
+        if (count++ > 0)
+          strcat(s, ", ");
+        strcat(s, error_flag[i]);
+      }
+  }
+
+  return s;
+}
+
+void ataPrintDriveInfo (struct ata_identify_device *drive){
   int version, drivetype;
   const char *description;
   char unknown[64], timedatetz[64];
@@ -112,13 +248,13 @@ void ataPrintDriveInfo (struct hd_driveid *drive){
 
   // print out model, serial # and firmware versions  (byte-swap ASCI strings)
   pout("Device Model:     ");
-  printswap(model, drive->model,40);
+  printswap(model, (char *)drive->model,40);
 
   pout("Serial Number:    ");
-  printswap(serial, drive->serial_no,20);
+  printswap(serial, (char *)drive->serial_no,20);
 
   pout("Firmware Version: ");
-  printswap(firm, drive->fw_rev,8);
+  printswap(firm, (char *)drive->fw_rev,8);
 
   // See if drive is recognized
   drivetype=lookupdrive(model, firm);
@@ -172,7 +308,7 @@ void PrintSmartOfflineStatus(struct ata_smart_values *data){
   // the final 7 bits
   unsigned char stat=data->offline_data_collection_status & 0x7f;
   
-  pout("Off-line data collection status: (0x%02x)\t",
+  pout("Offline data collection status:  (0x%02x)\t",
        (int)data->offline_data_collection_status);
     
   switch(stat){
@@ -207,9 +343,9 @@ void PrintSmartOfflineStatus(struct ata_smart_values *data){
   // Report on Automatic Data Collection Status.  Only IBM documents
   // this bit.  See SFF 8035i Revision 2 for details.
   if (data->offline_data_collection_status & 0x80)
-    pout("\t\t\t\t\tAuto Off-line Data Collection: Enabled.\n");
+    pout("\t\t\t\t\tAuto Offline Data Collection: Enabled.\n");
   else
-    pout("\t\t\t\t\tAuto Off-line Data Collection: Disabled.\n");
+    pout("\t\t\t\t\tAuto Offline Data Collection: Disabled.\n");
   
   return;
 }
@@ -283,7 +419,7 @@ void PrintSmartSelfExecStatus(struct ata_smart_values *data)
 
 
 void PrintSmartTotalTimeCompleteOffline ( struct ata_smart_values *data){
-  pout("Total time to complete off-line \n");
+  pout("Total time to complete Offline \n");
   pout("data collection: \t\t (%4d) seconds.\n", 
        (int)data->total_time_to_complete_off_line);
 }
@@ -296,7 +432,7 @@ void PrintSmartOfflineCollectCap(struct ata_smart_values *data){
        (int)data->offline_data_collection_capability);
   
   if (data->offline_data_collection_capability == 0x00){
-    pout("\tOff-line data collection not supported.\n");
+    pout("\tOffline data collection not supported.\n");
   } 
   else {
     pout( "%s\n", isSupportExecuteOfflineImmediate(data)?
@@ -304,8 +440,8 @@ void PrintSmartOfflineCollectCap(struct ata_smart_values *data){
 	  "No SMART execute Offline immediate.");
     
     pout( "\t\t\t\t\t%s\n", isSupportAutomaticTimer(data)? 
-	  "Automatic timer ON/OFF support.":
-	  "No Automatic timer ON/OFF support.");
+	  "Auto Offline data collection on/off support.":
+	  "No Auto Offline data collection support.");
     
     pout( "\t\t\t\t\t%s\n", isSupportOfflineAbort(data)? 
 	  "Abort Offline collection upon new\n\t\t\t\t\tcommand.":
@@ -480,7 +616,7 @@ void PrintSmartAttribWithThres (struct ata_smart_values *data,
   if (!needheader) pout("\n");
 }
 
-void ataPrintGeneralSmartValues(struct ata_smart_values *data, struct hd_driveid *drive){
+void ataPrintGeneralSmartValues(struct ata_smart_values *data, struct ata_identify_device *drive){
   pout("General SMART Values:\n");
   
   PrintSmartOfflineStatus(data); 
@@ -594,7 +730,7 @@ int ataPrintSmartErrorlog(struct ata_smart_errorlog *data){
   }
 
   // Some internal consistency checking of the data structures
-  if ((data->ata_error_count-data->error_log_pointer)%5) {
+  if ((data->ata_error_count-data->error_log_pointer)%5 && con->fixfirmwarebug != FIX_SAMSUNG2) {
     pout("Warning: ATA error count %d inconsistent with error log pointer %d\n\n",
 	 data->ata_error_count,data->error_log_pointer);
   }
@@ -621,6 +757,7 @@ int ataPrintSmartErrorlog(struct ata_smart_errorlog *data){
   
   // now step through the five error log data structures (table 39 of spec)
   for (k = 4; k >= 0; k-- ) {
+    char *st_er_desc;
 
     // The error log data structure entries are a circular buffer
     i=(data->error_log_pointer+k)%5;
@@ -635,7 +772,7 @@ int ataPrintSmartErrorlog(struct ata_smart_errorlog *data){
       case 0x01: msgstate="sleeping"; break;
       case 0x02: msgstate="in standby mode"; break;
       case 0x03: msgstate="active or idle"; break;
-      case 0x04: msgstate="doing SMART off-line or self test"; break;
+      case 0x04: msgstate="doing SMART Offline or Self-test"; break;
       default:   
 	if (bits<0x0b)
 	  msgstate="in a reserved state";
@@ -652,7 +789,7 @@ int ataPrintSmartErrorlog(struct ata_smart_errorlog *data){
       pout("  After command completion occurred, registers were:\n"
 	   "  ER ST SC SN CL CH DH\n"
 	   "  -- -- -- -- -- -- --\n"
-	   "  %02x %02x %02x %02x %02x %02x %02x\n\n",
+	   "  %02x %02x %02x %02x %02x %02x %02x",
 	   (int)data->errorlog_struct[i].error_struct.error_register,
 	   (int)data->errorlog_struct[i].error_struct.status,
 	   (int)data->errorlog_struct[i].error_struct.sector_count,
@@ -660,6 +797,19 @@ int ataPrintSmartErrorlog(struct ata_smart_errorlog *data){
 	   (int)data->errorlog_struct[i].error_struct.cylinder_low,
 	   (int)data->errorlog_struct[i].error_struct.cylinder_high,
 	   (int)data->errorlog_struct[i].error_struct.drive_head);
+      // Add a description of the contents of the status and error registers
+      // if possible
+      st_er_desc = construct_st_er_desc(
+        data->errorlog_struct[i].commands[4].commandreg,
+        data->errorlog_struct[i].commands[4].featuresreg,
+        data->errorlog_struct[i].error_struct.status,
+        data->errorlog_struct[i].error_struct.error_register
+      );
+      if (st_er_desc) {
+        pout("  %s", st_er_desc);
+        free(st_er_desc);
+      }
+      pout("\n\n");
       pout("  Commands leading to the command that caused the error were:\n"
 	   "  CR FR SC SN CL CH DH DC   Timestamp  Command/Feature_Name\n"
 	   "  -- -- -- -- -- -- -- --   ---------  --------------------\n");
@@ -691,10 +841,13 @@ int ataPrintSmartErrorlog(struct ata_smart_errorlog *data){
   return data->ata_error_count;  
 }
 
-// return value is number of entries found where the self-test showed an error
+// return value is:
+// bottom 8 bits: number of entries found where self-test showed an error
+// remaining bits: if nonzero, power on hours of last self-test where error was found
 int ataPrintSmartSelfTestlog(struct ata_smart_selftestlog *data,int allentries){
   int i,j,noheaderprinted=1;
-  int retval=0;
+  int retval=0, hours=0;
+
 
   if (allentries)
     pout("SMART Self-test log structure revision number %d\n",(int)data->revnumber);
@@ -702,7 +855,7 @@ int ataPrintSmartSelfTestlog(struct ata_smart_selftestlog *data,int allentries){
     pout("Warning: ATA Specification requires self-test log structure revision number = 1\n");
   if (data->mostrecenttest==0){
     if (allentries)
-      pout("No self-tests have been logged\n\n");
+      pout("No self-tests have been logged.  [Use the smartctl -t option to run these.]\n\n");
     return 0;
   }
 
@@ -719,12 +872,12 @@ int ataPrintSmartSelfTestlog(struct ata_smart_selftestlog *data,int allentries){
 
       // test name
       switch(log->selftestnumber){
-      case   0: msgtest="Off-line           "; break;
-      case   1: msgtest="Short off-line     "; break;
-      case   2: msgtest="Extended off-line  "; break;
-      case   3: msgtest="Conveyance off-line"; break;
-      case   4: msgtest="Selective off-line "; break;
-      case 127: msgtest="Abort off-line test"; break;
+      case   0: msgtest="Offline            "; break;
+      case   1: msgtest="Short offline      "; break;
+      case   2: msgtest="Extended offline   "; break;
+      case   3: msgtest="Conveyance offline "; break;
+      case   4: msgtest="Selective offline  "; break;
+      case 127: msgtest="Abort offline test "; break;
       case 129: msgtest="Short captive      "; break;
       case 130: msgtest="Extended captive   "; break;
       case 131: msgtest="Conveyance captive "; break;
@@ -732,9 +885,9 @@ int ataPrintSmartSelfTestlog(struct ata_smart_selftestlog *data,int allentries){
       default:  
 	if ( log->selftestnumber>=192 ||
 	    (log->selftestnumber>= 64 && log->selftestnumber<=126))
-	  msgtest="Vendor off-line    ";
+	  msgtest="Vendor offline     ";
 	else
-	  msgtest="Reserved off-line  ";
+	  msgtest="Reserved offline   ";
       }
       
       // test status
@@ -772,19 +925,27 @@ int ataPrintSmartSelfTestlog(struct ata_smart_selftestlog *data,int allentries){
       else	
 	sprintf(firstlba,"0x%08x",log->lbafirstfailure);
 
+      // print out a header if needed
       if (noheaderprinted && (allentries || errorfound)){
 	pout("Num  Test_Description    Status                  Remaining  LifeTime(hours)  LBA_of_first_error\n");
 	noheaderprinted=0;
       }
       
+      // print out an entry, either if we are printing all entries OR
+      // if an error was found
       if (allentries || errorfound)
-	pout("#%2d  %s %s %s  %8d         %s\n",21-i,msgtest,msgstat,
-	     percent,(int)log->timestamp,firstlba);
+	pout("#%2d  %s %s %s  %8d         %s\n",21-i,msgtest,msgstat, percent,(int)log->timestamp,firstlba);
+
+      // keep track of time of most recent error
+      if (errorfound && !hours)
+	hours=log->timestamp;
     }
   }
   if (!allentries && retval)
     pout("\n");
-  return retval;
+
+  hours = hours << 8;
+  return (retval | hours);
 }
 
 void ataPseudoCheckSmart ( struct ata_smart_values *data, 
@@ -815,22 +976,22 @@ void failuretest(int type, int returnvalue){
   // If this is an error in an "optional" SMART command
   if (type==OPTIONAL_CMD){
     if (con->conservative){
-      pout("An optional SMART command has failed: exiting.  To continue, set the tolerance level to something other than 'conservative'\n");
-      exit(returnvalue);
+      pout("An optional SMART command failed: exiting.  Remove '-T conservative' option to continue.\n");
+      EXIT(returnvalue);
     }
     return;
   }
 
   // If this is an error in a "mandatory" SMART command
   if (type==MANDATORY_CMD){
-    if (con->permissive)
+    if (con->permissive--)
       return;
-    pout("A mandatory SMART command has failed: exiting. To continue, use the -T option to set the tolerance level to 'permissive'\n");
-    exit(returnvalue);
+    pout("A mandatory SMART command failed: exiting. To continue, add one or more '-T permissive' options.\n");
+    EXIT(returnvalue);
   }
 
-  pout("Smartctl internal error in failuretest(type=%d). Please contact %s\n",type,PROJECTHOME);
-  exit(returnvalue|FAILCMD);
+  pout("Smartctl internal error in failuretest(type=%d). Please contact developers at %s\n",type,PROJECTHOME);
+  EXIT(returnvalue|FAILCMD);
 }
 
 // Used to warn users about invalid checksums.  Action to be taken may be
@@ -844,13 +1005,13 @@ void checksumwarning(const char *string){
 
   // user has asked us to fail on checksum errors
   if (con->checksumfail)
-    exit(FAILSMART);
+    EXIT(FAILSMART);
 
   return;
 }
 
 // Initialize to zero just in case some SMART routines don't work
-struct hd_driveid drive;
+struct ata_identify_device drive;
 struct ata_smart_values smartval;
 struct ata_smart_thresholds smartthres;
 struct ata_smart_errorlog smarterror;
@@ -858,18 +1019,18 @@ struct ata_smart_selftestlog smartselftest;
 
 int ataPrintMain (int fd){
   int timewait,code;
-  int returnval=0;
+  int returnval=0, retid=0, supported=0, needupdate=0;
 
   // Start by getting Drive ID information.  We need this, to know if SMART is supported.
-  if (ataReadHDIdentity(fd,&drive)){
-    pout("Smartctl: Hard Drive Read Identity Failed\n\n");
+  if ((retid=ataReadHDIdentity(fd,&drive))<0){
+    pout("Smartctl: Device Read Identity Failed (not an ATA/ATAPI device)\n\n");
     failuretest(MANDATORY_CMD, returnval|=FAILID);
   }
 
   // If requested, show which presets would be used for this drive and exit.
   if (con->showpresets) {
     showpresets(&drive);
-    exit(0);
+    EXIT(0);
   }
 
   // Use preset vendor attribute options unless user has requested otherwise.
@@ -879,7 +1040,7 @@ int ataPrintMain (int fd){
       applypresets(&drive, &charptr, con);
     else {
       pout("Fatal internal error in ataPrintMain()\n");
-      exit(1);
+      EXIT(returnval|=FAILCMD);
     }
   }
 
@@ -888,31 +1049,63 @@ int ataPrintMain (int fd){
     pout("=== START OF INFORMATION SECTION ===\n");
     ataPrintDriveInfo(&drive);
   }
-  
-  // now check if drive supports SMART; otherwise time to exit
-  if (!ataSmartSupport(&drive)){
-    pout("SMART support is: Unavailable - device lacks SMART capability.\n");
+
+  // Was this a packet device?
+  if (retid>0){
+    pout("SMART support is: Unavailable - Packet Interface Devices [this device: %s] don't support ATA SMART\n", packetdevicetype(retid-1));
     failuretest(MANDATORY_CMD, returnval|=FAILSMART);
-    pout("                  Checking to be sure by trying SMART ENABLE command.\n");
+  }
+  
+  // if drive does not supports SMART it's time to exit
+  supported=ataSmartSupport(&drive);
+  if (supported != 1){
+    if (supported==0) {
+      pout("SMART support is: Unavailable - device lacks SMART capability.\n");
+      failuretest(MANDATORY_CMD, returnval|=FAILSMART);
+      pout("                  Checking to be sure by trying SMART ENABLE command.\n");
+    }
+    else {
+      pout("SMART support is: Ambiguous - ATA IDENTIFY DEVICE words 82-83 don't show if SMART supported.\n");
+      failuretest(MANDATORY_CMD, returnval|=FAILSMART);
+      pout("                  Checking for SMART support by trying SMART ENABLE command.\n");
+    }
+
     if (ataEnableSmart(fd)){
-      pout("                  No SMART functionality found. Sorry.\n");
-      failuretest(MANDATORY_CMD,returnval|=FAILSMART);
+      pout("                  SMART ENABLE failed - this establishes that this device lacks SMART functionality.\n");
+      failuretest(MANDATORY_CMD, returnval|=FAILSMART);
+      supported=0;
+    }
+    else {
+      pout("                  SMART ENABLE appeared to work!  Continuing.\n");
+      supported=1;
     }
-    else
-      pout("                  SMART appears to work.  Continuing.\n"); 
     if (!con->driveinfo) pout("\n");
   }
   
   // Now print remaining drive info: is SMART enabled?    
   if (con->driveinfo){
-    pout("SMART support is: Available - device has SMART capability.\n");
-    if (ataDoesSmartWork(fd))
-      pout("SMART support is: Enabled\n");
+    int ison=ataIsSmartEnabled(&drive),isenabled=ison;
+    
+    if (ison==-1) {
+      pout("SMART support is: Ambiguous - ATA IDENTIFY DEVICE words 85-87 don't show if SMART is enabled.\n");
+      failuretest(MANDATORY_CMD, returnval|=FAILSMART);
+      // check SMART support by trying a command
+      pout("                  Checking to be sure by trying SMART RETURN STATUS command.\n");
+      isenabled=ataDoesSmartWork(fd);
+    }
     else
-      pout("SMART support is: Disabled\n");
+      pout("SMART support is: Available - device has SMART capability.\n");
+    
+    if (isenabled)
+      pout("SMART support is: Enabled\n");
+    else {
+      if (ison==-1)
+	pout("SMART support is: Unavailable\n");
+      else
+	pout("SMART support is: Disabled\n");
+    }
     pout("\n");
   }
-
   
   // START OF THE ENABLE/DISABLE SECTION OF THE CODE
   if (con->smartenable || con->smartdisable || 
@@ -985,6 +1178,7 @@ int ataPrintMain (int fd){
       pout("Warning: device does not support SMART Automatic Timers.\n\n");
       failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
     }
+    needupdate=1;
     if (ataEnableAutoOffline(fd)){
       pout( "Smartctl: SMART Enable Automatic Offline Failed.\n\n");
       failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
@@ -997,6 +1191,7 @@ int ataPrintMain (int fd){
       pout("Warning: device does not support SMART Automatic Timers.\n\n");
       failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
     }
+    needupdate=1;
     if (ataDisableAutoOffline(fd)){
       pout("Smartctl: SMART Disable Automatic Offline Failed.\n\n");
       failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
@@ -1005,6 +1200,11 @@ int ataPrintMain (int fd){
       pout("SMART Automatic Offline Testing Disabled.\n");
   }
 
+  if (needupdate && ataReadSmartValues(fd, &smartval)){
+    pout("Smartctl: SMART Read Values failed.\n\n");
+    failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
+  }
+
   // all this for a newline!
   if (con->smartenable || con->smartdisable || 
       con->smartautosaveenable || con->smartautosavedisable || 
@@ -1187,7 +1387,7 @@ int ataPrintMain (int fd){
   switch (con->testcase){
   case OFFLINE_FULL_SCAN:
     if (!isSupportExecuteOfflineImmediate(&smartval)){
-      pout("Warning: device does not support Execute Off-Line Immediate function.\n\n");
+      pout("Warning: device does not support Execute Offline Immediate function.\n\n");
       failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
     }
     break;
@@ -1219,8 +1419,8 @@ int ataPrintMain (int fd){
 #endif
   default:
     pout("Internal error in smartctl: con->testcase==%d not recognized\n", (int)con->testcase);
-    pout("Please contact smartmontools developers.\n");
-    exit(returnval|=FAILCMD);
+    pout("Please contact smartmontools developers at %s.\n", PACKAGE_BUGREPORT);
+    EXIT(returnval|=FAILCMD);
   }
 
   // Now do the test.  Note ataSmartTest prints its own error/success
@@ -1244,8 +1444,15 @@ int ataPrintMain (int fd){
   
   // Now say how long the test will take to complete
   if ((timewait=TestTime(&smartval,con->testcase))){ 
-    pout("Please wait %d %s for test to complete.\n",
-	 (int)timewait, con->testcase==OFFLINE_FULL_SCAN?"seconds":"minutes");
+    time_t t=time(NULL);
+    if (con->testcase==OFFLINE_FULL_SCAN) {
+      t+=timewait;
+      pout("Please wait %d seconds for test to complete.\n", (int)timewait);
+    } else {
+      t+=timewait*60;
+      pout("Please wait %d minutes for test to complete.\n", (int)timewait);
+    }
+    pout("Test will complete after %s\n", ctime(&t));
     
     if (con->testcase!=SHORT_CAPTIVE_SELF_TEST && 
 	con->testcase!=EXTEND_CAPTIVE_SELF_TEST && 
