@@ -40,20 +40,19 @@
 
 #include <stdio.h>
 #include <string.h>
-#include <getopt.h>
 #include <unistd.h>
 #include <sys/ioctl.h>
 #include "scsicmds.h"
 #include "utility.h"
 #include "extern.h"
 
-const char *scsicmds_c_cvsid="$Id: scsicmds.c,v 1.51 2003/08/18 12:37:41 dpgilbert Exp $" EXTERN_H_CVSID SCSICMDS_H_CVSID;
+const char *scsicmds_c_cvsid="$Id: scsicmds.c,v 1.65 2003/11/20 01:02:27 dpgilbert Exp $" EXTERN_H_CVSID SCSICMDS_H_CVSID;
 
 /* for passing global control variables */
 extern smartmonctrl *con;
 
 /* output binary in hex and optionally ascii */
-static void dStrHex(const char* str, int len, int no_ascii)
+void dStrHex(const char* str, int len, int no_ascii)
 {
     const char* p = str;
     unsigned char c;
@@ -140,158 +139,11 @@ const char * scsi_get_opcode_name(UINT8 opcode)
     return NULL;
 }
 
-/* SCSI command transmission interface function, implementation is OS
- * specific. Returns 0 if SCSI command successfully launched and response
- * received, else returns a negative errno value */
-static int do_scsi_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop);
-
-/* <<<<<<<<<<<<<<<< Start of Linux specific code >>>>>>>>>>>>>>>>> */
-#if 1   
-/* Linux specific code, FreeBSD could conditionally compile in CAM stuff 
- * instead of this. */
-
-/* #include <scsi/scsi.h>       bypass for now */
-/* #include <scsi/scsi_ioctl.h> bypass for now */
-
-#define MAX_DXFER_LEN 1024      /* can be increased if necessary */
-#define SEND_IOCTL_RESP_SENSE_LEN 16    /* ioctl limitation */
-#define DRIVER_SENSE  0x8       /* alternate CHECK CONDITION indication */
-
-#ifndef SCSI_IOCTL_SEND_COMMAND
-#define SCSI_IOCTL_SEND_COMMAND 1
-#endif
-#ifndef SCSI_IOCTL_TEST_UNIT_READY
-#define SCSI_IOCTL_TEST_UNIT_READY 2
-#endif
-
-struct linux_ioctl_send_command
-{
-    int inbufsize;
-    int outbufsize;
-    UINT8 buff[MAX_DXFER_LEN + 16];
-};
-
-/* The Linux SCSI_IOCTL_SEND_COMMAND ioctl is primitive and it doesn't 
- * support: CDB length (guesses it from opcode), resid and timeout.
- * Patches in Linux 2.4.21 and 2.5.70 to extend SEND DIAGNOSTIC timeout
- * to 2 hours in order to allow long foreground extended self tests. */
-static int linux_do_scsi_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop)
-{
-    struct linux_ioctl_send_command wrk;
-    int status, buff_offset;
-    size_t len;
-
-    memcpy(wrk.buff, iop->cmnd, iop->cmnd_len);
-    buff_offset = iop->cmnd_len;
-    if (con->reportscsiioctl > 0) {
-        int k;
-        const unsigned char * ucp = iop->cmnd;
-        const char * np;
-
-        np = scsi_get_opcode_name(ucp[0]);
-        pout(" [%s: ", np ? np : "<unknown opcode>");
-        for (k = 0; k < iop->cmnd_len; ++k)
-            pout("%02x ", ucp[k]);
-        if ((con->reportscsiioctl > 1) && 
-            (DXFER_TO_DEVICE == iop->dxfer_dir) && (iop->dxferp)) {
-            int trunc = (iop->dxfer_len > 256) ? 1 : 0;
-
-            pout("]\n  Outgoing data, len=%d%s:\n", (int)iop->dxfer_len,
-                 (trunc ? " [only first 256 bytes shown]" : ""));
-            dStrHex(iop->dxferp, (trunc ? 256 : iop->dxfer_len) , 1);
-        }
-        else
-            pout("]");
-    }
-    switch (iop->dxfer_dir) {
-        case DXFER_NONE:
-            wrk.inbufsize = 0;
-            wrk.outbufsize = 0;
-            break;
-        case DXFER_FROM_DEVICE:
-            wrk.inbufsize = 0;
-            if (iop->dxfer_len > MAX_DXFER_LEN)
-                return -EINVAL;
-            wrk.outbufsize = iop->dxfer_len;
-            break;
-        case DXFER_TO_DEVICE:
-            if (iop->dxfer_len > MAX_DXFER_LEN)
-                return -EINVAL;
-            memcpy(wrk.buff + buff_offset, iop->dxferp, iop->dxfer_len);
-            wrk.inbufsize = iop->dxfer_len;
-            wrk.outbufsize = 0;
-            break;
-        default:
-            pout("do_scsi_cmnd_io: bad dxfer_dir\n");
-            return -EINVAL;
-    }
-    iop->resp_sense_len = 0;
-    iop->scsi_status = 0;
-    iop->resid = 0;
-    status = ioctl(dev_fd, SCSI_IOCTL_SEND_COMMAND , &wrk);
-    if (-1 == status) {
-        if (con->reportscsiioctl)
-            pout("  status=-1, errno=%d [%s]\n", errno, strerror(errno));
-        return -errno;
-    }
-    if (0 == status) {
-        if (con->reportscsiioctl > 0)
-            pout("  status=0\n");
-        if (DXFER_FROM_DEVICE == iop->dxfer_dir) {
-            memcpy(iop->dxferp, wrk.buff, iop->dxfer_len);
-            if (con->reportscsiioctl > 1) {
-                int trunc = (iop->dxfer_len > 256) ? 1 : 0;
-
-                pout("  Incoming data, len=%d%s:\n", (int)iop->dxfer_len,
-                     (trunc ? " [only first 256 bytes shown]" : ""));
-                dStrHex(iop->dxferp, (trunc ? 256 : iop->dxfer_len) , 1);
-            }
-        }
-        return 0;
-    }
-    iop->scsi_status = status & 0x7e; /* bits 0 and 7 used to be for vendors */
-    if (DRIVER_SENSE == ((status >> 24) & 0xf))
-        iop->scsi_status = SCSI_STATUS_CHECK_CONDITION;
-    len = (SEND_IOCTL_RESP_SENSE_LEN < iop->max_sense_len) ?
-                SEND_IOCTL_RESP_SENSE_LEN : iop->max_sense_len;
-    if ((SCSI_STATUS_CHECK_CONDITION == iop->scsi_status) && 
-        iop->sensep && (len > 0)) {
-        memcpy(iop->sensep, wrk.buff, len);
-        iop->resp_sense_len = len;
-        if (con->reportscsiioctl > 1) {
-            pout("  >>> Sense buffer, len=%d:\n", (int)len);
-            dStrHex(wrk.buff, len , 1);
-        }
-    }
-    if (con->reportscsiioctl) {
-        if (SCSI_STATUS_CHECK_CONDITION == iop->scsi_status) {
-            pout("  status=%x: sense_key=%x asc=%x ascq=%x\n", status & 0xff,
-                 wrk.buff[2] & 0xf, wrk.buff[12], wrk.buff[13]);
-        }
-        else
-            pout("  status=0x%x\n", status);
-    }
-    if (iop->scsi_status > 0)
-        return 0;
-    else {
-        if (con->reportscsiioctl > 0)
-            pout("  ioctl status=0x%x but scsi status=0, fail with EIO\n", 
-                 status);
-        return -EIO;      /* give up, assume no device there */
-    }
-}
-
-static int do_scsi_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop)
-{
-    return linux_do_scsi_cmnd_io(dev_fd, iop);
-}
-#endif
-/* <<<<<<<<<<<<<<<< End of Linux specific code >>>>>>>>>>>>>>>>> */
 
 void scsi_do_sense_disect(const struct scsi_cmnd_io * io_buf,
                           struct scsi_sense_disect * out)
 {
-    memset(out, 0, sizeof(out));
+    memset(out, 0, sizeof(struct scsi_sense_disect));
     if ((SCSI_STATUS_CHECK_CONDITION == io_buf->scsi_status) && 
         (io_buf->resp_sense_len > 7)) {  
         out->error_code = (io_buf->sensep[0] & 0x7f);
@@ -305,9 +157,17 @@ void scsi_do_sense_disect(const struct scsi_cmnd_io * io_buf,
 
 static int scsiSimpleSenseFilter(const struct scsi_sense_disect * sinfo)
 {
-    if (SCSI_SK_NOT_READY == sinfo->sense_key)
-        return SIMPLE_ERR_NOT_READY;
-    else if (SCSI_SK_ILLEGAL_REQUEST == sinfo->sense_key) {
+    if (SCSI_SK_NOT_READY == sinfo->sense_key) {
+        if (SCSI_ASC_NO_MEDIUM == sinfo->asc) 
+            return SIMPLE_ERR_NO_MEDIUM;
+        else if (SCSI_ASC_NOT_READY == sinfo->asc) {
+            if (0x1 == sinfo->ascq)
+                return SIMPLE_ERR_BECOMING_READY;
+            else
+                return SIMPLE_ERR_NOT_READY;
+        } else
+            return SIMPLE_ERR_NOT_READY;
+    } else if (SCSI_SK_ILLEGAL_REQUEST == sinfo->sense_key) {
         if (SCSI_ASC_UNKNOWN_OPCODE == sinfo->asc)
             return SIMPLE_ERR_BAD_OPCODE;
         else if (SCSI_ASC_UNKNOWN_FIELD == sinfo->asc)
@@ -335,6 +195,10 @@ const char * scsiErrString(int scsiErr)
             return "badly formed scsi parameters";
         case SIMPLE_ERR_BAD_RESP: 
             return "scsi response fails sanity test";
+        case SIMPLE_ERR_NO_MEDIUM: 
+            return "no medium present";
+        case SIMPLE_ERR_BECOMING_READY: 
+            return "device will be ready soon";
         default:
             return "unknown error";
     }
@@ -369,6 +233,8 @@ int scsiLogSense(int device, int pagenum, UINT8 *pBuf, int bufLen,
         pageLen = 4;
         if (pageLen > bufLen)
             return -EIO;
+        else
+            memset(pBuf, 0, pageLen);
 
         memset(&io_hdr, 0, sizeof(io_hdr));
         memset(cdb, 0, sizeof(cdb));
@@ -384,16 +250,22 @@ int scsiLogSense(int device, int pagenum, UINT8 *pBuf, int bufLen,
         io_hdr.sensep = sense;
         io_hdr.max_sense_len = sizeof(sense);
     
-        status = do_scsi_cmnd_io(device, &io_hdr);
+        status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
         if (0 != status)
             return status;
         scsi_do_sense_disect(&io_hdr, &sinfo);
         if ((res = scsiSimpleSenseFilter(&sinfo)))
             return res;
+        /* sanity check on response */
+        if ((SUPPORTED_LPAGES != pagenum) && (pBuf[0] != pagenum))
+            return SIMPLE_ERR_BAD_RESP;
+        if (0 == ((pBuf[2] << 8) + pBuf[3]))
+            return SIMPLE_ERR_BAD_RESP;
         pageLen = (pBuf[2] << 8) + pBuf[3] + 4;
         if (pageLen > bufLen)
             pageLen = bufLen;
     }
+    memset(pBuf, 0, 4);
     memset(&io_hdr, 0, sizeof(io_hdr));
     memset(cdb, 0, sizeof(cdb));
     io_hdr.dxfer_dir = DXFER_FROM_DEVICE;
@@ -408,11 +280,19 @@ int scsiLogSense(int device, int pagenum, UINT8 *pBuf, int bufLen,
     io_hdr.sensep = sense;
     io_hdr.max_sense_len = sizeof(sense);
 
-    status = do_scsi_cmnd_io(device, &io_hdr);
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
         return status;
     scsi_do_sense_disect(&io_hdr, &sinfo);
-    return scsiSimpleSenseFilter(&sinfo);
+    status = scsiSimpleSenseFilter(&sinfo);
+    if (0 != status)
+        return status;
+    /* sanity check on response */
+    if ((SUPPORTED_LPAGES != pagenum) && (pBuf[0] != pagenum))
+        return SIMPLE_ERR_BAD_RESP;
+    if (0 == ((pBuf[2] << 8) + pBuf[3]))
+        return SIMPLE_ERR_BAD_RESP;
+    return 0;
 }
 
 /* Send MODE SENSE (6 byte) command. Returns 0 if ok, 1 if NOT READY,
@@ -442,7 +322,7 @@ int scsiModeSense(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen)
     io_hdr.sensep = sense;
     io_hdr.max_sense_len = sizeof(sense);
 
-    status = do_scsi_cmnd_io(device, &io_hdr);
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 == status) {
         scsi_do_sense_disect(&io_hdr, &sinfo);
         status = scsiSimpleSenseFilter(&sinfo);
@@ -466,7 +346,7 @@ int scsiModeSense(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen)
  * 2 if command not supported (then MODE SELECT(10) may be supported), 
  * 3 if field in command not supported, 4 if bad parameter to command
  * or returns negated errno. SPC sections 7.7 and 8.4 */
-int scsiModeSelect(int device, int pagenum, int sp, UINT8 *pBuf, int bufLen)
+int scsiModeSelect(int device, int sp, UINT8 *pBuf, int bufLen)
 {
     struct scsi_cmnd_io io_hdr;
     struct scsi_sense_disect sinfo;
@@ -496,7 +376,7 @@ int scsiModeSelect(int device, int pagenum, int sp, UINT8 *pBuf, int bufLen)
     io_hdr.sensep = sense;
     io_hdr.max_sense_len = sizeof(sense);
 
-    status = do_scsi_cmnd_io(device, &io_hdr);
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
         return status;
     scsi_do_sense_disect(&io_hdr, &sinfo);
@@ -529,7 +409,7 @@ int scsiModeSense10(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen)
     io_hdr.sensep = sense;
     io_hdr.max_sense_len = sizeof(sense);
 
-    status = do_scsi_cmnd_io(device, &io_hdr);
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 == status) {
         scsi_do_sense_disect(&io_hdr, &sinfo);
         status = scsiSimpleSenseFilter(&sinfo);
@@ -553,7 +433,7 @@ int scsiModeSense10(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen)
  * command not supported (then MODE SELECT(6) may be supported), 3 if field
  * in command not supported, 4 if bad parameter to command or returns
  * negated errno. SAM sections 7.8 and 8.4 */
-int scsiModeSelect10(int device, int pagenum, int sp, UINT8 *pBuf, int bufLen)
+int scsiModeSelect10(int device, int sp, UINT8 *pBuf, int bufLen)
 {
     struct scsi_cmnd_io io_hdr;
     struct scsi_sense_disect sinfo;
@@ -584,7 +464,7 @@ int scsiModeSelect10(int device, int pagenum, int sp, UINT8 *pBuf, int bufLen)
     io_hdr.sensep = sense;
     io_hdr.max_sense_len = sizeof(sense);
 
-    status = do_scsi_cmnd_io(device, &io_hdr);
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
         return status;
     scsi_do_sense_disect(&io_hdr, &sinfo);
@@ -612,7 +492,7 @@ int scsiStdInquiry(int device, UINT8 *pBuf, int bufLen)
     cdb[4] = bufLen;
     io_hdr.cmnd = cdb;
     io_hdr.cmnd_len = sizeof(cdb);
-    status = do_scsi_cmnd_io(device, &io_hdr);
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
         return status;
     scsi_do_sense_disect(&io_hdr, &sinfo);
@@ -649,7 +529,7 @@ int scsiInquiryVpd(int device, int vpd_page, UINT8 *pBuf, int bufLen)
     io_hdr.sensep = sense;
     io_hdr.max_sense_len = sizeof(sense);
 
-    status = do_scsi_cmnd_io(device, &io_hdr);
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
         return status;
     scsi_do_sense_disect(&io_hdr, &sinfo);
@@ -685,7 +565,7 @@ int scsiRequestSense(int device, struct scsi_sense_disect * sense_info)
     cdb[4] = sizeof(buff);
     io_hdr.cmnd = cdb;
     io_hdr.cmnd_len = sizeof(cdb);
-    status = do_scsi_cmnd_io(device, &io_hdr);
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if ((0 == status) && (sense_info)) {
         ecode = buff[0] & 0x7f;
         sense_info->error_code = ecode;
@@ -735,7 +615,7 @@ int scsiSendDiagnostic(int device, int functioncode, UINT8 *pBuf, int bufLen)
     io_hdr.timeout = 5 * 60 * 60;   /* five hours because a foreground 
                     extended self tests can take 1 hour plus */
     
-    status = do_scsi_cmnd_io(device, &io_hdr);
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
         return status;
     scsi_do_sense_disect(&io_hdr, &sinfo);
@@ -769,7 +649,7 @@ int scsiReceiveDiagnostic(int device, int pcv, int pagenum, UINT8 *pBuf,
     io_hdr.sensep = sense;
     io_hdr.max_sense_len = sizeof(sense);
 
-    status = do_scsi_cmnd_io(device, &io_hdr);
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
         return status;
     scsi_do_sense_disect(&io_hdr, &sinfo);
@@ -795,7 +675,7 @@ static int _testunitready(int device, struct scsi_sense_disect * sinfo)
     io_hdr.sensep = sense;
     io_hdr.max_sense_len = sizeof(sense);
 
-    status = do_scsi_cmnd_io(device, &io_hdr);
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
         return status;
     scsi_do_sense_disect(&io_hdr, sinfo);
@@ -812,28 +692,25 @@ int scsiTestUnitReady(int device)
     status = _testunitready(device, &sinfo);
     if (0 != status)
         return status;
-    if (SCSI_SK_NOT_READY == sinfo.sense_key)
-        return SIMPLE_ERR_NOT_READY;
-    else if (SCSI_SK_UNIT_ATTENTION == sinfo.sense_key) {
+
+    if (SCSI_SK_UNIT_ATTENTION == sinfo.sense_key) {
         /* power on reset, media changed, ok ... try again */
         status = _testunitready(device, &sinfo);        
         if (0 != status)
             return status;
-        if (SCSI_SK_NOT_READY == sinfo.sense_key)
-            return SIMPLE_ERR_BAD_FIELD;
     }
-    return 0;
+    return scsiSimpleSenseFilter(&sinfo);
 }
 
 /* Offset into mode sense (6 or 10 byte) response that actual mode page
  * starts at (relative to resp[0]). Returns -1 if problem */
-int scsiModePageOffset(const UINT8 * resp, int len, int modese_10)
+int scsiModePageOffset(const UINT8 * resp, int len, int modese_len)
 {
     int resp_len, bd_len;
     int offset = -1;
 
     if (resp) {
-        if (modese_10) {
+        if (10 == modese_len) {
             resp_len = (resp[0] << 8) + resp[1] + 2;
             bd_len = (resp[6] << 8) + resp[7];
             offset = bd_len + 8;
@@ -867,82 +744,58 @@ int scsiModePageOffset(const UINT8 * resp, int len, int modese_10)
  * tries a 10 byte MODE SENSE command. Returns 0 if successful, a positive
  * number if a known error (see  SIMPLE_ERR_ ...) or a negative errno
  * value. */
-int scsiFetchIECmpage(int device, struct scsi_iec_mode_page *iecp)
+int scsiFetchIECmpage(int device, struct scsi_iec_mode_page *iecp, int modese_len)
 {
-    int err;
+    int err = 0;
 
     memset(iecp, 0, sizeof(*iecp));
+    iecp->modese_len = modese_len;
     iecp->requestedCurrent = 1;
-    if ((err = scsiModeSense(device, INFORMATIONAL_EXCEPTIONS_CONTROL, 
-                             0, iecp->raw_curr, sizeof(iecp->raw_curr)))) {
-        if (SIMPLE_ERR_BAD_OPCODE == err) { /* so try 10 byte mode sense */
-            err = scsiModeSense10(device, INFORMATIONAL_EXCEPTIONS_CONTROL, 
-                             0, iecp->raw_curr, sizeof(iecp->raw_curr));
-            if (0 == err)
-                iecp->modese_10 = 1;
-            else
+    if (iecp->modese_len <= 6) {
+        if ((err = scsiModeSense(device, INFORMATIONAL_EXCEPTIONS_CONTROL_PAGE, 
+                                 MPAGE_CONTROL_CURRENT, 
+                                 iecp->raw_curr, sizeof(iecp->raw_curr)))) {
+            if (SIMPLE_ERR_BAD_OPCODE == err)
+                iecp->modese_len = 10;
+            else {
+                iecp->modese_len = 0;
                 return err;
-        } else
+            }
+        } else if (0 == iecp->modese_len)
+            iecp->modese_len = 6;
+    }
+    if (10 == iecp->modese_len) {
+        err = scsiModeSense10(device, INFORMATIONAL_EXCEPTIONS_CONTROL_PAGE,
+                              MPAGE_CONTROL_CURRENT, 
+                              iecp->raw_curr, sizeof(iecp->raw_curr));
+        if (err) {
+            iecp->modese_len = 0;
             return err;
+        }
     } 
     iecp->gotCurrent = 1;
     iecp->requestedChangeable = 1;
-    if (iecp->modese_10) {
-        if (0 == scsiModeSense10(device, INFORMATIONAL_EXCEPTIONS_CONTROL, 
-                                 1, iecp->raw_chg, sizeof(iecp->raw_chg)))
-            iecp->gotChangeable = 1;
-    } else {
-        if (0 == scsiModeSense(device, INFORMATIONAL_EXCEPTIONS_CONTROL, 
-                               1, iecp->raw_chg, sizeof(iecp->raw_chg)))
-            iecp->gotChangeable = 1;
-    }
+    if (10 == iecp->modese_len)
+        err = scsiModeSense10(device, INFORMATIONAL_EXCEPTIONS_CONTROL_PAGE,
+                                 MPAGE_CONTROL_CHANGEABLE,
+                                 iecp->raw_chg, sizeof(iecp->raw_chg));
+    else if (6 == iecp->modese_len)
+        err = scsiModeSense(device, INFORMATIONAL_EXCEPTIONS_CONTROL_PAGE, 
+                            MPAGE_CONTROL_CHANGEABLE, 
+                            iecp->raw_chg, sizeof(iecp->raw_chg));
+    if (err)
+        return err;
+    iecp->gotChangeable = 1;
     return 0;
 }
 
-/* Return 0 if ok, -EINVAL if problems */
-int scsiDecodeIEModePage(const struct scsi_iec_mode_page *iecp,
-                UINT8 *byte_2p, UINT8 *mrie_p, unsigned int *interval_timer_p,
-                unsigned int *report_count_p)
-{
-    int offset, len;
-
-    if (iecp && iecp->gotCurrent) {
-        offset = scsiModePageOffset(iecp->raw_curr, sizeof(iecp->raw_curr),
-                                    iecp->modese_10);
-        if (offset >= 0) {
-            len = iecp->raw_curr[offset + 1] + 2;
-            if (byte_2p)
-                *byte_2p = iecp->raw_curr[offset + 2];
-            if (mrie_p)
-                *mrie_p = iecp->raw_curr[offset + 3] & 0xf;
-            if (interval_timer_p && (len > 7))
-                *interval_timer_p = (iecp->raw_curr[offset + 4] << 24) +
-                                    (iecp->raw_curr[offset + 5] << 16) +
-                                    (iecp->raw_curr[offset + 6] << 8) +
-                                    iecp->raw_curr[offset + 7];
-            else if (interval_timer_p)
-                *interval_timer_p = 0;
-            if (report_count_p && (len > 11))
-                *report_count_p = (iecp->raw_curr[offset + 8] << 24) +
-                                  (iecp->raw_curr[offset + 9] << 16) +
-                                  (iecp->raw_curr[offset + 10] << 8) +
-                                  iecp->raw_curr[offset + 11];
-            else if (report_count_p)
-                *report_count_p = 0;
-            return 0;
-        } else
-            return -EINVAL;
-    } else
-        return -EINVAL;
-}
-
 int scsi_IsExceptionControlEnabled(const struct scsi_iec_mode_page *iecp)
 {
     int offset;
 
     if (iecp && iecp->gotCurrent) {
         offset = scsiModePageOffset(iecp->raw_curr, sizeof(iecp->raw_curr),
-                                    iecp->modese_10);
+                                    iecp->modese_len);
         if (offset >= 0)
             return (iecp->raw_curr[offset + 2] & DEXCPT_ENABLE) ? 0 : 1;
         else
@@ -957,7 +810,7 @@ int scsi_IsWarningEnabled(const struct scsi_iec_mode_page *iecp)
 
     if (iecp && iecp->gotCurrent) {
         offset = scsiModePageOffset(iecp->raw_curr, sizeof(iecp->raw_curr),
-                                    iecp->modese_10);
+                                    iecp->modese_len);
         if (offset >= 0)
             return (iecp->raw_curr[offset + 2] & EWASC_ENABLE) ? 1 : 0;
         else
@@ -984,20 +837,25 @@ int scsi_IsWarningEnabled(const struct scsi_iec_mode_page *iecp)
 int scsiSetExceptionControlAndWarning(int device, int enabled,
                                       const struct scsi_iec_mode_page *iecp)
 {
-    int k, offset, err;
+    int k, offset, resp_len;
+    int err = 0;
     UINT8 rout[SCSI_IECMP_RAW_LEN];
     int sp, eCEnabled, wEnabled;
 
     if ((! iecp) || (! iecp->gotCurrent))
         return -EINVAL;
     offset = scsiModePageOffset(iecp->raw_curr, sizeof(iecp->raw_curr),
-                                iecp->modese_10);
+                                iecp->modese_len);
     if (offset < 0)
         return -EINVAL;
     memcpy(rout, iecp->raw_curr, SCSI_IECMP_RAW_LEN);
-    rout[0] = 0;     /* Mode Data Length reserved in MODE SELECTs */
-    if (iecp->modese_10)
-        rout[1] = 0;
+    if (10 == iecp->modese_len) {
+            resp_len = (rout[0] << 8) + rout[1] + 2;
+            memset(rout, 0, 2); /* mode data length==0 for mode select */
+    } else {
+            resp_len = rout[0] + 1;
+            memset(rout, 0, 1); /* mode data length==0 for mode select */
+    }
     sp = (rout[offset] & 0x80) ? 1 : 0; /* PS bit becomes 'SELECT's SP bit */
     rout[offset] &= 0x7f;     /* mask off PS bit */
     if (enabled) {
@@ -1045,12 +903,10 @@ int scsiSetExceptionControlAndWarning(int device, int enabled,
                 rout[offset + 2] &= TEST_DISABLE;/* clear TEST bit for spec */
         }
     }
-    if (iecp->modese_10)
-        err = scsiModeSelect10(device, INFORMATIONAL_EXCEPTIONS_CONTROL, 
-                               sp, rout, sizeof(rout));
-    else
-        err = scsiModeSelect(device, INFORMATIONAL_EXCEPTIONS_CONTROL, 
-                             sp, rout, sizeof(rout));
+    if (10 == iecp->modese_len)
+        err = scsiModeSelect10(device, sp, rout, resp_len);
+    else if (6 == iecp->modese_len)
+        err = scsiModeSelect(device, sp, rout, resp_len);
     return err;
 }
 
@@ -1059,7 +915,8 @@ int scsiGetTemp(int device, UINT8 *currenttemp, UINT8 *triptemp)
     UINT8 tBuf[252];
     int err;
 
-    if ((err = scsiLogSense(device, TEMPERATURE_PAGE, tBuf, 
+    memset(tBuf, 0, sizeof(tBuf));
+    if ((err = scsiLogSense(device, TEMPERATURE_LPAGE, tBuf, 
                             sizeof(tBuf), 0))) {
         *currenttemp = 0;
         *triptemp = 0;
@@ -1076,26 +933,30 @@ int scsiGetTemp(int device, UINT8 *currenttemp, UINT8 *triptemp)
  * Returns 0 if ok, else error number. A current temperature of 255
  * (Celsius) implies that the temperature not available. */
 int scsiCheckIE(int device, int hasIELogPage, int hasTempLogPage,
-                UINT8 *asc, UINT8 *ascq, UINT8 *currenttemp)
+                UINT8 *asc, UINT8 *ascq, UINT8 *currenttemp,
+                UINT8 *triptemp)
 {
     UINT8 tBuf[252];
     struct scsi_sense_disect sense_info;
     int err;
     int temperatureSet = 0;
     unsigned short pagesize;
-    UINT8 currTemp, tripTemp;
+    UINT8 currTemp, trTemp;
  
     *asc = 0;
     *ascq = 0;
     *currenttemp = 0;
+    *triptemp = 0;
+    memset(tBuf,0,sizeof(tBuf)); // need to clear stack space of junk
     memset(&sense_info, 0, sizeof(sense_info));
     if (hasIELogPage) {
-        if ((err = scsiLogSense(device, IE_LOG_PAGE, tBuf, 
+        if ((err = scsiLogSense(device, IE_LPAGE, tBuf, 
                                 sizeof(tBuf), 0))) {
             pout("Log Sense failed, IE page [%s]\n", scsiErrString(err));
             return err;
         }
-        pagesize = (unsigned short) (tBuf[2] << 8) | tBuf[3];
+	// pull out page size from response, don't forget to add 4
+        pagesize = (unsigned short) ((tBuf[2] << 8) | tBuf[3]) + 4; 
         if ((pagesize < 4) || tBuf[4] || tBuf[5]) {
             pout("Log Sense failed, IE page, bad parameter code or length\n");
             return SIMPLE_ERR_BAD_PARAM;
@@ -1103,9 +964,11 @@ int scsiCheckIE(int device, int hasIELogPage, int hasTempLogPage,
         if (tBuf[7] > 1) {
             sense_info.asc = tBuf[8]; 
             sense_info.ascq = tBuf[9];
-            if (tBuf[7] > 2) { 
-                *currenttemp = tBuf[10];
-                temperatureSet = 1;
+            if (! hasTempLogPage) {
+                if (tBuf[7] > 2) 
+                    *currenttemp = tBuf[10];
+                if (tBuf[7] > 3)        /* IBM extension in SMART (IE) lpage */
+                    *triptemp = tBuf[11];
             }
         } 
     }
@@ -1119,8 +982,10 @@ int scsiCheckIE(int device, int hasIELogPage, int hasTempLogPage,
     *asc = sense_info.asc;
     *ascq = sense_info.ascq;
     if ((! temperatureSet) && hasTempLogPage) {
-        if (0 == scsiGetTemp(device, &currTemp, &tripTemp))
+        if (0 == scsiGetTemp(device, &currTemp, &trTemp)) {
             *currenttemp = currTemp;
+            *triptemp = trTemp;
+        }
     }
     return 0;
 }
@@ -1699,26 +1564,33 @@ int scsiSmartSelfTestAbort(int device)
     return scsiSendDiagnostic(device, SCSI_DIAG_ABORT_SELF_TEST, NULL, 0);
 }
 
-int scsiFetchExtendedSelfTestTime(int device, int * durationSec)
+/* Returns 0 and the expected duration of an extended self test (in seconds)
+   if successful; any other return value indicates a failure. */
+int scsiFetchExtendedSelfTestTime(int device, int * durationSec, int modese_len)
 {
     int err, offset, res;
     UINT8 buff[64];
-    int modese_10 = 0;
 
     memset(buff, 0, sizeof(buff));
-    if ((err = scsiModeSense(device, CONTROL_MODE_PAGE_PARAMETERS, 
-                             0, buff, sizeof(buff)))) {
-        if (2 == err) { /* opcode no good so try 10 byte mode sense */
-            err = scsiModeSense10(device, CONTROL_MODE_PAGE_PARAMETERS, 
-                             0, buff, sizeof(buff));
-            if (0 == err)
-                modese_10 = 1;
+    if (modese_len <= 6) {
+        if ((err = scsiModeSense(device, CONTROL_MODE_PAGE, 
+                                 MPAGE_CONTROL_CURRENT, 
+                                 buff, sizeof(buff)))) {
+            if (SIMPLE_ERR_BAD_OPCODE == err)
+                modese_len = 10;
             else
                 return err;
-        } else
+        } else if (0 == modese_len)
+            modese_len = 6;
+    }
+    if (10 == modese_len) {
+        err = scsiModeSense10(device, CONTROL_MODE_PAGE, 
+                              MPAGE_CONTROL_CURRENT, 
+                              buff, sizeof(buff));
+        if (err)
             return err;
     } 
-    offset = scsiModePageOffset(buff, sizeof(buff), modese_10);
+    offset = scsiModePageOffset(buff, sizeof(buff), modese_len);
     if (offset < 0)
         return -EINVAL;
     if (buff[offset + 1] >= 0xa) {
@@ -1762,7 +1634,7 @@ void scsiDecodeErrCounterPage(unsigned char * resp,
 	}
 	k = pl - 4;
 	xp = ucp + 4;
-	if (k > sizeof(*ullp)) {
+	if (k > (int)sizeof(*ullp)) {
 	    xp += (k - sizeof(*ullp));
 	    k = sizeof(*ullp);
 	}
@@ -1815,3 +1687,196 @@ void scsiDecodeNonMediumErrPage(unsigned char *resp,
 	ucp += pl;
     }
 }
+
+/* Counts number of failed self-tests. Also encodes the poweron_hour
+   of the most recent failed self-test. Return value is negative if
+   this function has a problem (typically -1), otherwise the bottom 8
+   bits are the number of failed self tests and the 16 bits above that
+   are the poweron hour of the most recent failure. Note: aborted self
+   tests (typically by the user) and self tests in progress are not 
+   considered failures. See Working Draft SCSI Primary Commands - 3 
+   (SPC-3) section 7.2.10 T10/1416-D Rev 15 */
+int scsiCountFailedSelfTests(int fd, int noisy)
+{
+    int num, k, n, err, res, fails, fail_hour;
+    UINT8 * ucp;
+    unsigned char resp[LOG_RESP_SELF_TEST_LEN];
+
+    if ((err = scsiLogSense(fd, SELFTEST_RESULTS_LPAGE, resp, 
+                            LOG_RESP_SELF_TEST_LEN, 0))) {
+        if (noisy)
+            pout("scsiCountSelfTests Failed [%s]\n", scsiErrString(err));
+        return -1;
+    }
+    if (resp[0] != SELFTEST_RESULTS_LPAGE) {
+        if (noisy)
+            pout("Self-test Log Sense Failed, page mismatch\n");
+        return -1;
+    }
+    // compute page length
+    num = (resp[2] << 8) + resp[3];
+    // Log sense page length 0x190 bytes
+    if (num != 0x190) {
+        if (noisy)
+            pout("Self-test Log Sense length is 0x%x not 0x190 bytes\n", num);
+        return -1;
+    }
+    fails = 0;
+    fail_hour = 0;
+    // loop through the twenty possible entries
+    for (k = 0, ucp = resp + 4; k < 20; ++k, ucp += 20 ) {
+
+        // timestamp in power-on hours (or zero if test in progress)
+        n = (ucp[6] << 8) | ucp[7];
+
+        // The spec says "all 20 bytes will be zero if no test" but
+        // DG has found otherwise.  So this is a heuristic.
+        if ((0 == n) && (0 == ucp[4]))
+            break;
+        res = ucp[4] & 0xf;
+        if ((res > 2) && (res < 8)) {
+            fails++;
+            if (1 == fails) 
+                fail_hour = (ucp[6] << 8) + ucp[7];
+        }
+    }
+    return (fail_hour << 8) + fails;
+}
+
+/* Returns a negative value if failed to fetch Contol mode page or it was
+   malformed. Returns 0 if GLTSD bit is zero and returns 1 if the GLTSD
+   bit is set. Examines default mode page when current==0 else examines
+   current mode page. */
+int scsiFetchControlGLTSD(int device, int modese_len, int current)
+{
+    int err, offset;
+    UINT8 buff[64];
+    int pc = current ? MPAGE_CONTROL_CURRENT : MPAGE_CONTROL_DEFAULT;
+
+    memset(buff, 0, sizeof(buff));
+    if (modese_len <= 6) {
+        if ((err = scsiModeSense(device, CONTROL_MODE_PAGE, pc, 
+                                 buff, sizeof(buff)))) {
+            if (SIMPLE_ERR_BAD_OPCODE == err)
+                modese_len = 10;
+            else
+                return -EINVAL;
+        } else if (0 == modese_len)
+            modese_len = 6;
+    }
+    if (10 == modese_len) {
+        err = scsiModeSense10(device, CONTROL_MODE_PAGE, pc,
+                              buff, sizeof(buff));
+        if (err)
+            return -EINVAL;
+    } 
+    offset = scsiModePageOffset(buff, sizeof(buff), modese_len);
+    if ((offset >= 0) && (buff[offset + 1] >= 0xa))
+        return (buff[offset + 2] & 2) ? 1 : 0;
+    return -EINVAL;
+}
+
+/* Attempts to set or clear GLTSD bit in Control mode page. If enabled is
+   0 attempts to clear GLTSD otherwise it attempts to set it. Returns 0 if
+   successful, negative if low level error, > 0 if higher level error (e.g.
+   SIMPLE_ERR_BAD_PARAM if GLTSD bit is not changeable). */
+int scsiSetControlGLTSD(int device, int enabled, int modese_len)
+{
+    int err, offset, resp_len, sp;
+    UINT8 buff[64];
+    UINT8 ch_buff[64];
+
+    memset(buff, 0, sizeof(buff));
+    if (modese_len <= 6) {
+        if ((err = scsiModeSense(device, CONTROL_MODE_PAGE, 
+                                 MPAGE_CONTROL_CURRENT, 
+                                 buff, sizeof(buff)))) {
+            if (SIMPLE_ERR_BAD_OPCODE == err)
+                modese_len = 10;
+            else
+                return err;
+        } else if (0 == modese_len)
+            modese_len = 6;
+    }
+    if (10 == modese_len) {
+        err = scsiModeSense10(device, CONTROL_MODE_PAGE, 
+                              MPAGE_CONTROL_CURRENT, 
+                              buff, sizeof(buff));
+        if (err)
+            return err;
+    } 
+    offset = scsiModePageOffset(buff, sizeof(buff), modese_len);
+    if ((offset < 0) || (buff[offset + 1] < 0xa))
+        return SIMPLE_ERR_BAD_RESP;
+
+    if (enabled)
+        enabled = 2;
+    if (enabled == (buff[offset + 2] & 2))
+        return 0;       /* GLTSD already in wanted state so nothing to do */
+
+    if (modese_len == 6)
+        err = scsiModeSense(device, CONTROL_MODE_PAGE, 
+                            MPAGE_CONTROL_CHANGEABLE, 
+                            ch_buff, sizeof(ch_buff));
+    else
+        err = scsiModeSense10(device, CONTROL_MODE_PAGE, 
+                              MPAGE_CONTROL_CHANGEABLE, 
+                              ch_buff, sizeof(ch_buff));
+    if (err)
+        return err;
+    if (0 == (ch_buff[offset + 2] & 2))
+        return SIMPLE_ERR_BAD_PARAM;  /* GLTSD bit not chageable */
+    
+    if (10 == modese_len) {
+            resp_len = (buff[0] << 8) + buff[1] + 2;
+            memset(buff, 0, 2);
+    } else {
+            resp_len = buff[0] + 1;
+            memset(buff, 0, 1);
+    }
+    sp = (buff[offset] & 0x80) ? 1 : 0; /* PS bit becomes 'SELECT's SP bit */
+    buff[offset] &= 0x7f;     /* mask off PS bit */
+    if (enabled)
+        buff[offset + 2] |= 0x2;    /* set GLTSD bit */
+    else
+        buff[offset + 2] &= 0xfd;   /* clear GLTSD bit */
+    if (10 == modese_len)
+        err = scsiModeSelect10(device, sp, buff, resp_len);
+    else if (6 == modese_len)
+        err = scsiModeSelect(device, sp, buff, resp_len);
+    return err;
+}
+
+/* Returns a negative value if failed to fetch Protocol specific port mode 
+   page or it was malformed. Returns transport protocol identifier when
+   value >= 0 . */
+int scsiFetchTransportProtocol(int device, int modese_len)
+{
+    int err, offset;
+    UINT8 buff[64];
+
+    memset(buff, 0, sizeof(buff));
+    if (modese_len <= 6) {
+        if ((err = scsiModeSense(device, PROTOCOL_SPECIFIC_PORT_PAGE, 
+                                 MPAGE_CONTROL_CURRENT, 
+                                 buff, sizeof(buff)))) {
+            if (SIMPLE_ERR_BAD_OPCODE == err)
+                modese_len = 10;
+            else
+                return -EINVAL;
+        } else if (0 == modese_len)
+            modese_len = 6;
+    }
+    if (10 == modese_len) {
+        err = scsiModeSense10(device, PROTOCOL_SPECIFIC_PORT_PAGE, 
+                              MPAGE_CONTROL_CURRENT, 
+                              buff, sizeof(buff));
+        if (err)
+            return -EINVAL;
+    } 
+    offset = scsiModePageOffset(buff, sizeof(buff), modese_len);
+    if ((offset >= 0) && (buff[offset + 1] > 1))
+        return (buff[offset + 2] & 0xf);
+    return -EINVAL;
+}
+
