@@ -32,14 +32,11 @@
 #ifndef SCSICMDS_H_
 #define SCSICMDS_H_
 
-#ifndef SCSICMDS_H_CVSID
-#define SCSICMDS_H_CVSID "$Id: scsicmds.h,v 1.31 2003/08/18 12:36:36 dpgilbert Exp $\n"
-#endif
+#define SCSICMDS_H_CVSID "$Id: scsicmds.h,v 1.43 2003/11/20 00:34:09 dpgilbert Exp $\n"
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <getopt.h>
 #include <unistd.h>
 #include <errno.h>
 #include <sys/ioctl.h>
@@ -123,7 +120,7 @@ struct scsi_iec_mode_page {
     UINT8 gotCurrent;
     UINT8 requestedChangeable;
     UINT8 gotChangeable;
-    UINT8 modese_10;
+    UINT8 modese_len;   /* 0 (don't know), 6 or 10 */
     UINT8 raw_curr[SCSI_IECMP_RAW_LEN];
     UINT8 raw_chg[SCSI_IECMP_RAW_LEN];
 };
@@ -150,62 +147,64 @@ struct scsiNonMediumError {
 #define SCSI_PT_ENCLOSURE               0xd
 
 /* ANSI SCSI-3 Log Pages retrieved by LOG SENSE. */
-#define SUPPORTED_LOG_PAGES                     0x00
-#define BUFFER_OVERRUN_PAGE                     0x01
-#define WRITE_ERROR_COUNTER_PAGE                0x02
-#define READ_ERROR_COUNTER_PAGE                 0x03
-#define READ_REVERSE_ERROR_COUNTER_PAGE         0x04
-#define VERIFY_ERROR_COUNTER_PAGE               0x05
-#define NON_MEDIUM_ERROR_PAGE                   0x06
-#define LAST_N_ERROR_PAGE                       0x07
-#define FORMAT_STATUS_PAGE                      0x08
-#define TEMPERATURE_PAGE                        0x0d
-#define STARTSTOP_CYCLE_COUNTER_PAGE            0x0e
-#define APPLICATION_CLIENT_PAGE                 0x0f
-#define SELFTEST_RESULTS_PAGE                   0x10
-#define IE_LOG_PAGE                             0x2f
+#define SUPPORTED_LPAGES                        0x00
+#define BUFFER_OVERRUN_LPAGE                    0x01
+#define WRITE_ERROR_COUNTER_LPAGE               0x02
+#define READ_ERROR_COUNTER_LPAGE                0x03
+#define READ_REVERSE_ERROR_COUNTER_LPAGE        0x04
+#define VERIFY_ERROR_COUNTER_LPAGE              0x05
+#define NON_MEDIUM_ERROR_LPAGE                  0x06
+#define LAST_N_ERROR_LPAGE                      0x07
+#define FORMAT_STATUS_LPAGE                     0x08
+#define TEMPERATURE_LPAGE                       0x0d
+#define STARTSTOP_CYCLE_COUNTER_LPAGE           0x0e
+#define APPLICATION_CLIENT_LPAGE                0x0f
+#define SELFTEST_RESULTS_LPAGE                  0x10
+#define IE_LPAGE                                0x2f
+
+/* Seagate vendor specific log pages. */
+#define SEAGATE_CACHE_LPAGE                     0x37
+#define SEAGATE_FACTORY_LPAGE                   0x3e
+
+/* Log page response lengths */
+#define LOG_RESP_SELF_TEST_LEN 0x194
 
 /* See the SSC-2 document at www.t10.org . Earler note: From IBM 
 Documentation, see http://www.storage.ibm.com/techsup/hddtech/prodspecs.htm */
-#define TAPE_ALERTS_PAGE                         0x2e
+#define TAPE_ALERTS_LPAGE                        0x2e
 
 /* ANSI SCSI-3 Mode Pages */
-#define VENDOR_UNIQUE_PARAMETERS                 0x00
-#define READ_WRITE_ERROR_RECOVERY_PARAMETERS     0x01
-#define DISCONNECT_RECONNECT_PARAMETERS          0x02
-#define FORMAT_DEVICE_PARAMETERS                 0x03
-#define RIGID_DISK_DRIVE_GEOMETRY_PARAMETERS     0x04
-#define FLEXIBLE_DISK_PARAMETERS                 0x05
-#define VERIFY_ERROR_RECOVERY_PARAMETERS         0x07
-#define CACHING_PARAMETERS                       0x08
-#define PERIPHERAL_DEVICE                        0x09
-#define XOR_CONTROL_MODE_PARAMETERS              0x10
-#define CONTROL_MODE_PAGE_PARAMETERS             0x0a
-#define MEDIUM_TYPES_SUPPORTED                   0x0b
-#define NOTCH_PARAMETERS                         0x0c
-#define POWER_CONDITION_PARAMETERS               0x0d
-#define CD_DEVICE_PARAMETERS                     0x0d
+#define VENDOR_UNIQUE_PAGE                       0x00
+#define READ_WRITE_ERROR_RECOVERY_PAGE           0x01
+#define DISCONNECT_RECONNECT_PAGE                0x02
+#define FORMAT_DEVICE_PAGE                       0x03
+#define RIGID_DISK_DRIVE_GEOMETRY_PAGE           0x04
+#define FLEXIBLE_DISK_PAGE                       0x05
+#define VERIFY_ERROR_RECOVERY_PAGE               0x07
+#define CACHING_PAGE                             0x08
+#define PERIPHERAL_DEVICE_PAGE                   0x09
+#define XOR_CONTROL_MODE_PAGE                    0x10
+#define CONTROL_MODE_PAGE                        0x0a
+#define MEDIUM_TYPES_SUPPORTED_PAGE              0x0b
+#define NOTCH_PAGE                               0x0c
+#define CD_DEVICE_PAGE                           0x0d
 #define CD_AUDIO_CONTROL_PAGE                    0x0e
-#define DATA_COMPRESSION_PARAMETERS              0x0f
-#define MEDIUM_PARTITION_MODE_PARAMETERS_1       0x11
-#define MEDIUM_PARTITION_MODE_PARAMETERS_2       0x12
-#define MEDIUM_PARTITION_MODE_PARAMETERS_3       0x13
-#define MEDIUM_PARTITION_MODE_PARAMETERS_4       0x14
-#define ENCLOSURE_SERVICES_MANAGEMENT            0x14
-#define LUN_CONTROL                              0x18
-#define PORT_CONTROL                             0x19
-#define POWER_CONTROL                            0x1a
-#define LUN_MAPPING_PAGE                         0x1b
-#define INFORMATIONAL_EXCEPTIONS_CONTROL         0x1c
+#define DATA_COMPRESSION_PAGE                    0x0f
+#define ENCLOSURE_SERVICES_MANAGEMENT_PAGE       0x14
+#define PROTOCOL_SPECIFIC_LUN_PAGE               0x18
+#define PROTOCOL_SPECIFIC_PORT_PAGE              0x19
+#define POWER_CONDITION_PAGE                     0x1a
+#define INFORMATIONAL_EXCEPTIONS_CONTROL_PAGE    0x1c
 #define FAULT_FAILURE_REPORTING_PAGE             0x1c
-#define ELEMENT_ADDRESS_ASSIGNMENT               0x1d
-#define TIMEOUT_AND_PROTECT_PARAMETERS           0x1d
-#define TRANSPORT_GEOMETRY_PARAMETERS            0x1e
-#define DEVICE_CAPABILITIES                      0x1f
-#define CD_CAPABILITIES_AND_MECHANISM_STATUS     0x2a
 
 #define ALL_MODE_PAGES                           0x3f
 
+/* Mode page control field */
+#define MPAGE_CONTROL_CURRENT               0
+#define MPAGE_CONTROL_CHANGEABLE            1
+#define MPAGE_CONTROL_DEFAULT               2
+#define MPAGE_CONTROL_SAVED                 3
+
 /* defines for useful SCSI Status codes */
 #define SCSI_STATUS_CHECK_CONDITION     0x2
 
@@ -215,6 +214,8 @@ Documentation, see http://www.storage.ibm.com/techsup/hddtech/prodspecs.htm */
 #define SCSI_SK_UNIT_ATTENTION          0x6
 
 /* defines for useful Additional Sense Codes (ASCs) */
+#define SCSI_ASC_NOT_READY              0x4     /* more info in ASCQ code */
+#define SCSI_ASC_NO_MEDIUM              0x3a    /* more info in ASCQ code */
 #define SCSI_ASC_UNKNOWN_OPCODE         0x20
 #define SCSI_ASC_UNKNOWN_FIELD          0x24
 #define SCSI_ASC_UNKNOWN_PARAM          0x26
@@ -228,6 +229,8 @@ Documentation, see http://www.storage.ibm.com/techsup/hddtech/prodspecs.htm */
 #define SIMPLE_ERR_BAD_FIELD            3       /* in cbd */
 #define SIMPLE_ERR_BAD_PARAM            4       /* in data */
 #define SIMPLE_ERR_BAD_RESP             5       /* response fails sanity */
+#define SIMPLE_ERR_NO_MEDIUM            6       /* no medium present */
+#define SIMPLE_ERR_BECOMING_READY       7       /* device will be ready soon */
 
 
 /* defines for functioncode parameter in SENDDIAGNOSTIC function */
@@ -261,13 +264,13 @@ int scsiLogSense(int device, int pagenum, UINT8 *pBuf, int bufLen,
 
 int scsiModeSense(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen);
 
-int scsiModeSelect(int device, int pagenum, int sp, UINT8 *pBuf, int bufLen);
+int scsiModeSelect(int device, int sp, UINT8 *pBuf, int bufLen);
 
 int scsiModeSense10(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen);
 
-int scsiModeSelect10(int device, int pagenum, int sp, UINT8 *pBuf, int bufLen);
+int scsiModeSelect10(int device, int sp, UINT8 *pBuf, int bufLen);
 
-int scsiModePageOffset(const UINT8 * resp, int len, int modese_10);
+int scsiModePageOffset(const UINT8 * resp, int len, int modese_len);
 
 int scsiRequestSense(int device, struct scsi_sense_disect * sense_info);
 
@@ -277,22 +280,25 @@ int scsiReceiveDiagnostic(int device, int pcv, int pagenum, UINT8 *pBuf,
                       int bufLen);
 
 /* SMART specific commands */
-int scsiCheckIE(int device, int hasIELogPage, int hasTempLogPage,
-                UINT8 *asc, UINT8 *ascq, UINT8 *currenttemp);
+int scsiCheckIE(int device, int hasIELogPage, int hasTempLogPage, UINT8 *asc,
+                UINT8 *ascq, UINT8 *currenttemp, UINT8 *triptemp);
 
-int scsiFetchIECmpage(int device, struct scsi_iec_mode_page *iecp);
+int scsiFetchIECmpage(int device, struct scsi_iec_mode_page *iecp, 
+                      int modese_len);
 int scsi_IsExceptionControlEnabled(const struct scsi_iec_mode_page *iecp);
 int scsi_IsWarningEnabled(const struct scsi_iec_mode_page *iecp);
 int scsiSetExceptionControlAndWarning(int device, int enabled,
                             const struct scsi_iec_mode_page *iecp);
-int scsiDecodeIEModePage(const struct scsi_iec_mode_page *iecp,
-        UINT8 *byte_2p, UINT8 *mrie_p, unsigned int *interval_timer_p,
-        unsigned int *report_count_p);
 void scsiDecodeErrCounterPage(unsigned char * resp,
                               struct scsiErrorCounter *ecp);
 void scsiDecodeNonMediumErrPage(unsigned char * resp,
                                 struct scsiNonMediumError *nmep);
-int scsiFetchExtendedSelfTestTime(int device, int * durationSec);
+int scsiFetchExtendedSelfTestTime(int device, int * durationSec, 
+                                  int modese_len);
+int scsiCountFailedSelfTests(int fd, int noisy);
+int scsiFetchControlGLTSD(int device, int modese_len, int current);
+int scsiSetControlGLTSD(int device, int enabled, int modese_len);
+int scsiFetchTransportProtocol(int device, int modese_len);
 
 /* T10 Standard IE Additional Sense Code strings taken from t10.org */
 
@@ -311,5 +317,11 @@ int scsiSmartSelfTestAbort(int device);
 
 const char * scsiTapeAlertsTapeDevice(unsigned short code);
 const char * scsiTapeAlertsChangerDevice(unsigned short code);
+
+const char * scsi_get_opcode_name(UINT8 opcode);
+void dStrHex(const char* str, int len, int no_ascii);
+
+int do_scsi_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop, int report);
+
 #endif
 
