@@ -40,16 +40,18 @@
 
 #define GBUF_SIZE 65535
 
-const char* scsiprint_c_cvsid="$Id: scsiprint.c,v 1.49 2003/06/17 06:10:08 dpgilbert Exp $"
+const char* scsiprint_c_cvsid="$Id: scsiprint.c,v 1.70 2003/11/24 10:55:33 dpgilbert Exp $"
 EXTERN_H_CVSID SCSICMDS_H_CVSID SCSIPRINT_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
 
 // control block which points to external global control variables
 extern smartmonctrl *con;
 
+// to hold onto exit code for atexit routine
+extern int exitstatus;
+
 UINT8 gBuf[GBUF_SIZE];
 #define LOG_RESP_LEN 252
 #define LOG_RESP_TAPE_ALERT_LEN 0x144
-#define LOG_RESP_SELF_TEST_LEN 0x194
 
 /* Log pages supported */
 static int gSmartLPage = 0;     /* Informational Exceptions log page */
@@ -57,16 +59,24 @@ static int gTempLPage = 0;
 static int gSelfTestLPage = 0;
 static int gStartStopLPage = 0;
 static int gTapeAlertsLPage = 0;
+static int gSeagateCacheLPage = 0;
+static int gSeagateFactoryLPage = 0;
 
 /* Mode pages supported */
 static int gIecMPage = 1;     /* N.B. assume it until we know otherwise */
 
+/* Remember last successful mode sense/select command */
+static int modese_len = 0;
+
+// Compares failure type to policy in effect, and either exits or
+// simply returns to the calling routine.
+extern void failuretest(int type, int returnvalue);
 
 static void scsiGetSupportedLogPages(int device)
 {
     int i, err;
 
-    if ((err = scsiLogSense(device, SUPPORTED_LOG_PAGES, gBuf, 
+    if ((err = scsiLogSense(device, SUPPORTED_LPAGES, gBuf, 
                             LOG_RESP_LEN, 0))) {
         if (con->reportscsiioctl > 0)
             pout("Log Sense for supported pages failed [%s]\n", 
@@ -77,21 +87,27 @@ static void scsiGetSupportedLogPages(int device)
     for (i = 4; i < gBuf[3] + LOGPAGEHDRSIZE; i++) {
         switch (gBuf[i])
         {
-            case TEMPERATURE_PAGE:
+            case TEMPERATURE_LPAGE:
                 gTempLPage = 1;
                 break;
-            case STARTSTOP_CYCLE_COUNTER_PAGE:
+            case STARTSTOP_CYCLE_COUNTER_LPAGE:
                 gStartStopLPage = 1;
                 break;
-            case SELFTEST_RESULTS_PAGE:
+            case SELFTEST_RESULTS_LPAGE:
                 gSelfTestLPage = 1;
                 break;
-            case IE_LOG_PAGE:
+            case IE_LPAGE:
                 gSmartLPage = 1;
                 break;
-            case TAPE_ALERTS_PAGE:
+            case TAPE_ALERTS_LPAGE:
                 gTapeAlertsLPage = 1;
                 break;
+            case SEAGATE_CACHE_LPAGE:
+                gSeagateCacheLPage = 1;
+                break;
+            case SEAGATE_FACTORY_LPAGE:
+                gSeagateFactoryLPage = 1;
+                break;
             default:
                 break;
         }
@@ -103,12 +119,13 @@ void scsiGetSmartData(int device)
     UINT8 asc;
     UINT8 ascq;
     UINT8 currenttemp = 0;
+    UINT8 triptemp = 0;
     const char * cp;
     int err;
 
     QUIETON(con);
-    if ((err = scsiCheckIE(device, gSmartLPage, gTempLPage,
-                           &asc, &ascq, &currenttemp))) {
+    if ((err = scsiCheckIE(device, gSmartLPage, gTempLPage, &asc, 
+                           &ascq, &currenttemp, &triptemp))) {
         /* error message already announced */
         QUIETOFF(con);
         return;
@@ -128,6 +145,8 @@ void scsiGetSmartData(int device)
         else
             pout("Current Drive Temperature:     <not available>\n");
     }
+    if (triptemp && !gTempLPage)
+        pout("Drive Trip Temperature:        %d C\n", triptemp);
 }
 
 
@@ -145,7 +164,7 @@ static int scsiGetTapeAlertsData(int device, int peripheral_type)
     int failures = 0;
 
     QUIETON(con);
-    if ((err = scsiLogSense(device, TAPE_ALERTS_PAGE, gBuf, 
+    if ((err = scsiLogSense(device, TAPE_ALERTS_LPAGE, gBuf, 
                         LOG_RESP_TAPE_ALERT_LEN, LOG_RESP_TAPE_ALERT_LEN))) {
         pout("scsiGetTapesAlertData Failed [%s]\n", scsiErrString(err));
         QUIETOFF(con);
@@ -190,14 +209,14 @@ void scsiGetStartStopData(int device)
     int err, len, k;
     char str[6];
 
-    if ((err = scsiLogSense(device, STARTSTOP_CYCLE_COUNTER_PAGE, gBuf,
+    if ((err = scsiLogSense(device, STARTSTOP_CYCLE_COUNTER_LPAGE, gBuf,
                             LOG_RESP_LEN, 0))) {
         QUIETON(con);
         pout("scsiGetStartStopData Failed [%s]\n", scsiErrString(err));
         QUIETOFF(con);
         return;
     }
-    if (gBuf[0] != STARTSTOP_CYCLE_COUNTER_PAGE) {
+    if (gBuf[0] != STARTSTOP_CYCLE_COUNTER_LPAGE) {
         QUIETON(con);
         pout("StartStop Log Sense Failed, page mismatch\n");
         QUIETOFF(con);
@@ -220,11 +239,160 @@ void scsiGetStartStopData(int device)
         currentStartStop = (gBuf[36] << 24) | (gBuf[37] << 16) |
                            (gBuf[38] << 8) | gBuf[39];
         pout("Current start stop count:      %u times\n", currentStartStop);
-        pout("Recommended start stop count:  %u times\n", 
+        pout("Recommended maximum start stop count:  %u times\n", 
              recommendedStartStop);
     }
 } 
 
+static void scsiPrintSeagateCacheLPage(int device)
+{
+    int k, j, num, pl, pc, err, len;
+    unsigned char * ucp;
+    unsigned char * xp;
+    unsigned long long ull;
+
+    if ((err = scsiLogSense(device, SEAGATE_CACHE_LPAGE, gBuf,
+                            LOG_RESP_LEN, 0))) {
+        QUIETON(con);
+        pout("Seagate Cache Log Sense Failed: %s\n", scsiErrString(err));
+        QUIETOFF(con);
+        return;
+    }
+    if (gBuf[0] != SEAGATE_CACHE_LPAGE) {
+        QUIETON(con);
+        pout("Seagate Cache Log Sense Failed, page mismatch\n");
+        QUIETOFF(con);
+        return;
+    }
+    len = ((gBuf[2] << 8) | gBuf[3]) + 4;
+    num = len - 4;
+    ucp = &gBuf[0] + 4;
+    while (num > 3) {
+        pc = (ucp[0] << 8) | ucp[1];
+        pl = ucp[3] + 4;
+        switch (pc) {
+        case 0: case 1: case 2: case 3: case 4:
+            break;
+        default: 
+            if (con->reportscsiioctl > 0) {
+                QUIETON(con);
+                pout("\nVendor (Seagate) cache lpage has unexpected parameter"
+                     ", skip\n");
+                QUIETOFF(con);
+            }
+            return;
+        }
+        num -= pl;
+        ucp += pl;
+    }
+    pout("\nVendor (Seagate) cache information\n");
+    num = len - 4;
+    ucp = &gBuf[0] + 4;
+    while (num > 3) {
+        pc = (ucp[0] << 8) | ucp[1];
+        pl = ucp[3] + 4;
+        switch (pc) {
+        case 0: pout("  Blocks sent to initiator"); break;
+        case 1: pout("  Blocks received from initiator"); break;
+        case 2: pout("  Blocks read from cache and sent to initiator"); break;
+        case 3: pout("  Number of read and write commands whose size "
+                       "<= segment size"); break;
+        case 4: pout("  Number of read and write commands whose size "
+                       "> segment size"); break;
+        default: pout("  Unknown Seagate parameter code [0x%x]", pc); break;
+        }
+        k = pl - 4;
+        xp = ucp + 4;
+        if (k > (int)sizeof(ull)) {
+            xp += (k - (int)sizeof(ull));
+            k = (int)sizeof(ull);
+        }
+        ull = 0;
+        for (j = 0; j < k; ++j) {
+            if (j > 0)
+                ull <<= 8;
+            ull |= xp[j];
+        }
+        pout(" = %llu\n", ull);
+        num -= pl;
+        ucp += pl;
+    }
+}
+
+static void scsiPrintSeagateFactoryLPage(int device)
+{
+    int k, j, num, pl, pc, len, err;
+    unsigned char * ucp;
+    unsigned char * xp;
+    unsigned long long ull;
+
+    if ((err = scsiLogSense(device, SEAGATE_FACTORY_LPAGE, gBuf,
+                            LOG_RESP_LEN, 0))) {
+        QUIETON(con);
+        pout("scsiPrintSeagateFactoryLPage Failed [%s]\n", scsiErrString(err));
+        QUIETOFF(con);
+        return;
+    }
+    if (gBuf[0] != SEAGATE_FACTORY_LPAGE) {
+        QUIETON(con);
+        pout("Seagate Factory Log Sense Failed, page mismatch\n");
+        QUIETOFF(con);
+        return;
+    }
+    len = ((gBuf[2] << 8) | gBuf[3]) + 4;
+    num = len - 4;
+    ucp = &gBuf[0] + 4;
+    while (num > 3) {
+        pc = (ucp[0] << 8) | ucp[1];
+        pl = ucp[3] + 4;
+        switch (pc) {
+        case 0: case 8:
+            break;
+        default: 
+            if (con->reportscsiioctl > 0) {
+                QUIETON(con);
+                pout("\nVendor (Seagate) factory lpage has unexpected "
+                     "parameter, skip\n");
+                QUIETOFF(con);
+            }
+            return;
+        }
+        num -= pl;
+        ucp += pl;
+    }
+    pout("Vendor (Seagate) factory information\n");
+    num = len - 4;
+    ucp = &gBuf[0] + 4;
+    while (num > 3) {
+        pc = (ucp[0] << 8) | ucp[1];
+        pl = ucp[3] + 4;
+        switch (pc) {
+        case 0: pout("  number of hours powered up"); break;
+        case 8: pout("  number of minutes until next internal SMART test");
+            break;
+        default: pout("  Unknown Seagate parameter code [0x%x]", pc); break;
+        }
+        k = pl - 4;
+        xp = ucp + 4;
+        if (k > (int)sizeof(ull)) {
+            xp += (k - (int)sizeof(ull));
+            k = (int)sizeof(ull);
+        }
+        ull = 0;
+        for (j = 0; j < k; ++j) {
+            if (j > 0)
+                ull <<= 8;
+            ull |= xp[j];
+        }
+        if (0 == pc)
+            pout(" = %.2f\n", ((double)ull) / 60.0 );
+        else
+            pout(" = %llu\n", ull);
+        num -= pl;
+        ucp += pl;
+    }
+}
+
 static void scsiPrintErrorCounterLog(int device)
 {
     struct scsiErrorCounter errCounterArr[3];
@@ -235,17 +403,17 @@ static void scsiPrintErrorCounterLog(int device)
     int k;
     double processed_gb;
 
-    if (0 == scsiLogSense(device, READ_ERROR_COUNTER_PAGE, gBuf, 
+    if (0 == scsiLogSense(device, READ_ERROR_COUNTER_LPAGE, gBuf, 
                           LOG_RESP_LEN, 0)) {
         scsiDecodeErrCounterPage(gBuf, &errCounterArr[0]);
         found[0] = 1;
     }
-    if (0 == scsiLogSense(device, WRITE_ERROR_COUNTER_PAGE, gBuf, 
+    if (0 == scsiLogSense(device, WRITE_ERROR_COUNTER_LPAGE, gBuf, 
                           LOG_RESP_LEN, 0)) {
         scsiDecodeErrCounterPage(gBuf, &errCounterArr[1]);
         found[1] = 1;
     }
-    if (0 == scsiLogSense(device, VERIFY_ERROR_COUNTER_PAGE, gBuf, 
+    if (0 == scsiLogSense(device, VERIFY_ERROR_COUNTER_LPAGE, gBuf, 
                           LOG_RESP_LEN, 0)) {
         scsiDecodeErrCounterPage(gBuf, &errCounterArr[2]);
         ecp = &errCounterArr[2];
@@ -277,7 +445,7 @@ static void scsiPrintErrorCounterLog(int device)
     }
     else 
         pout("\nDevice does not support Error Counter logging\n");
-    if (0 == scsiLogSense(device, NON_MEDIUM_ERROR_PAGE, gBuf, 
+    if (0 == scsiLogSense(device, NON_MEDIUM_ERROR_LPAGE, gBuf, 
                           LOG_RESP_LEN, 0)) {
         scsiDecodeNonMediumErrPage(gBuf, &nme);
         if (nme.gotPC0)
@@ -324,14 +492,14 @@ static int scsiPrintSelfTest(int device)
     UINT8 * ucp;
     unsigned long long ull=0;
 
-    if ((err = scsiLogSense(device, SELFTEST_RESULTS_PAGE, gBuf, 
+    if ((err = scsiLogSense(device, SELFTEST_RESULTS_LPAGE, gBuf, 
                             LOG_RESP_SELF_TEST_LEN, 0))) {
         QUIETON(con);
         pout("scsiPrintSelfTest Failed [%s]\n", scsiErrString(err));
         QUIETOFF(con);
         return 1;
     }
-    if (gBuf[0] != SELFTEST_RESULTS_PAGE) {
+    if (gBuf[0] != SELFTEST_RESULTS_LPAGE) {
         QUIETON(con);
         pout("Self-test Log Sense Failed, page mismatch\n");
         QUIETOFF(con);
@@ -418,10 +586,11 @@ static int scsiPrintSelfTest(int device)
         pout("No self-tests have been logged\n");
     else
         pout("\n");
-    if ((0 == scsiFetchExtendedSelfTestTime(device, &durationSec)) &&
-        (durationSec > 0))
+    if ((0 == scsiFetchExtendedSelfTestTime(device, &durationSec, 
+                        modese_len)) && (durationSec > 0)) {
         pout("Long (extended) Self Test duration: %d seconds "
              "[%.1f minutes]\n", durationSec, durationSec / 60.0);
+    }
     return 0;
 }
 
@@ -443,6 +612,25 @@ static const char * peripheral_dt_arr[] = {
         "simplified disk",
         "optical card reader"
 };
+
+static const char * transport_proto_arr[] = {
+        "Fibre channel (FCP-2)",
+        "Parallel SCSI (SPI-4)",
+        "SSA",
+        "IEEE 1394 (SBP-2)",
+        "RDMA (SRP)",
+        "iSCSI",
+        "SAS",
+        "ADT",
+        "0x8",
+        "0x9",
+        "0xa",
+        "0xb",
+        "0xc",
+        "0xd",
+        "0xe",
+        "0xf"
+};
  
 /* Returns 0 on success */
 static int scsiGetDriveInfo(int device, UINT8 * peripheral_type, int all)
@@ -452,9 +640,10 @@ static int scsiGetDriveInfo(int device, UINT8 * peripheral_type, int all)
     char revision[5];
     char timedatetz[64];
     struct scsi_iec_mode_page iec;
-    int err, len;
+    int err, iec_err, len, val;
     int is_tape = 0;
     int peri_dt = 0;
+    int returnval=0;
         
     memset(gBuf, 0, 36);
     if ((err = scsiStdInquiry(device, gBuf, 36))) {
@@ -470,42 +659,63 @@ static int scsiGetDriveInfo(int device, UINT8 * peripheral_type, int all)
     if (! all)
 	return 0;
 
-    if (len >= 36) {
-        memset(manufacturer, 0, sizeof(manufacturer));
-        strncpy(manufacturer, &gBuf[8], 8);
-     
-        memset(product, 0, sizeof(product));
-        strncpy(product, &gBuf[16], 16);
-            
-        memset(revision, 0, sizeof(revision));
-        strncpy(revision, &gBuf[32], 4);
-        pout("Device: %s %s Version: %s\n", manufacturer, product, revision);
-        if (0 == (err = scsiInquiryVpd(device, 0x80, gBuf, 64))) {
-            /* should use VPD page 0x83 and fall back to this page (0x80)
-             * if 0x83 not supported. NAA requires a lot of decoding code */
-            len = gBuf[3];
-            gBuf[4 + len] = '\0';
-            pout("Serial number: %s\n", &gBuf[4]);
-        }
-        else if (con->reportscsiioctl > 0) {
-            QUIETON(con);
-            if (SIMPLE_ERR_BAD_RESP == err)
-                pout("Vital Product Data (VPD) bit ignored in INQUIRY\n");
-            else
-                pout("Vital Product Data (VPD) INQUIRY failed [%d]\n", err);
-            QUIETOFF(con);
-        }
-    } else {
+    if (len < 36) {
         QUIETON(con);
         pout("Short INQUIRY response, skip product id\n");
         QUIETOFF(con);
+        return 1;
     }
+    memset(manufacturer, 0, sizeof(manufacturer));
+    strncpy(manufacturer, (char *)&gBuf[8], 8);
+ 
+    memset(product, 0, sizeof(product));
+    strncpy(product, (char *)&gBuf[16], 16);
+        
+    memset(revision, 0, sizeof(revision));
+    strncpy(revision, (char *)&gBuf[32], 4);
+    pout("Device: %s %s Version: %s\n", manufacturer, product, revision);
+
+    /* Do this here to try and detect badly conforming devices (some USB
+       keys) that will lock up on a InquiryVpd or log sense or ... */
+    if ((iec_err = scsiFetchIECmpage(device, &iec, modese_len))) {
+        if (SIMPLE_ERR_BAD_RESP == iec_err) {
+            pout(">> Terminate command early due to bad response to IEC "
+                 "mode page\n");
+            QUIETOFF(con);
+            gIecMPage = 0;
+            return 1;
+        }
+    } else
+        modese_len = iec.modese_len;
+
+    if (0 == (err = scsiInquiryVpd(device, 0x80, gBuf, 64))) {
+        /* should use VPD page 0x83 and fall back to this page (0x80)
+         * if 0x83 not supported. NAA requires a lot of decoding code */
+        len = gBuf[3];
+        gBuf[4 + len] = '\0';
+        pout("Serial number: %s\n", &gBuf[4]);
+    }
+    else if (con->reportscsiioctl > 0) {
+        QUIETON(con);
+        if (SIMPLE_ERR_BAD_RESP == err)
+            pout("Vital Product Data (VPD) bit ignored in INQUIRY\n");
+        else
+            pout("Vital Product Data (VPD) INQUIRY failed [%d]\n", err);
+        QUIETOFF(con);
+    }
+
     // print SCSI peripheral device type
-    if (peri_dt < (sizeof(peripheral_dt_arr) / sizeof(peripheral_dt_arr[0])))
+    if (peri_dt < (int)(sizeof(peripheral_dt_arr) / 
+                        sizeof(peripheral_dt_arr[0])))
         pout("Device type: %s\n", peripheral_dt_arr[peri_dt]);
     else
         pout("Device type: <%d>\n", peri_dt);
 
+    // See if transport protocol is known
+    val = scsiFetchTransportProtocol(device, modese_len);
+    if ((val >= 0) && (val <= 0xf))
+        pout("Transport protocol: %s\n", transport_proto_arr[val]);
+
     // print current time and date and timezone
     dateandtimezone(timedatetz);
     pout("Local Time is: %s\n", timedatetz);
@@ -517,36 +727,38 @@ static int scsiGetDriveInfo(int device, UINT8 * peripheral_type, int all)
     if ((err = scsiTestUnitReady(device))) {
         if (SIMPLE_ERR_NOT_READY == err) {
             QUIETON(con);
-            pout("device is NOT READY (media absent, spun down, etc)\n");
+            pout("device is NOT READY (e.g. spun down, busy)\n");
+            QUIETOFF(con);
+         } else if (SIMPLE_ERR_NO_MEDIUM == err) {
+            QUIETON(con);
+            pout("NO MEDIUM present on device\n");
+            QUIETOFF(con);
+         } else if (SIMPLE_ERR_BECOMING_READY == err) {
+            QUIETON(con);
+            pout("device becoming ready (wait)\n");
             QUIETOFF(con);
         } else {
             QUIETON(con);
             pout("device Test Unit Ready  [%s]\n", scsiErrString(err));
             QUIETOFF(con);
         }
-	return 0;
+        failuretest(MANDATORY_CMD, returnval|=FAILID);
     }
    
-    if ((err = scsiFetchIECmpage(device, &iec))) {
+    if (iec_err) {
 	if (!is_tape) {
             QUIETON(con);
 	    pout("Device does not support SMART");
             if (con->reportscsiioctl > 0)
-	        pout(" [%s]\n", scsiErrString(err));
+	        pout(" [%s]\n", scsiErrString(iec_err));
             else
 	        pout("\n");
-            if (SIMPLE_ERR_BAD_RESP == err) {
-                pout(">> Terminate command early due to bad response to IEC "
-                     "mode page\n");
-                QUIETOFF(con);
-                gIecMPage = 0;
-                return 1;
-            }
             QUIETOFF(con);
         }
         gIecMPage = 0;
         return 0;
     }
+
     if (!is_tape)
         pout("Device supports SMART and is %s\n",
              (scsi_IsExceptionControlEnabled(&iec)) ? "Enabled" : "Disabled");
@@ -561,13 +773,15 @@ static int scsiSmartEnable(int device)
     struct scsi_iec_mode_page iec;
     int err;
 
-    if ((err = scsiFetchIECmpage(device, &iec))) {
+    if ((err = scsiFetchIECmpage(device, &iec, modese_len))) {
         QUIETON(con);
         pout("unable to fetch IEC (SMART) mode page [%s]\n", 
              scsiErrString(err));
         QUIETOFF(con);
         return 1;
-    }
+    } else
+        modese_len = iec.modese_len;
+
     if ((err = scsiSetExceptionControlAndWarning(device, 1, &iec))) {
         QUIETON(con);
         pout("unable to enable Exception control and warning [%s]\n",
@@ -576,11 +790,13 @@ static int scsiSmartEnable(int device)
         return 1;
     }
     /* Need to refetch 'iec' since could be modified by previous call */
-    if ((err = scsiFetchIECmpage(device, &iec))) {
+    if ((err = scsiFetchIECmpage(device, &iec, modese_len))) {
         pout("unable to fetch IEC (SMART) mode page [%s]\n", 
              scsiErrString(err));
         return 1;
-    }
+    } else
+        modese_len = iec.modese_len;
+
     pout("Informational Exceptions (SMART) %s\n",
          scsi_IsExceptionControlEnabled(&iec) ? "enabled" : "disabled");
     pout("Temperature warning %s\n",
@@ -593,13 +809,15 @@ static int scsiSmartDisable(int device)
     struct scsi_iec_mode_page iec;
     int err;
 
-    if ((err = scsiFetchIECmpage(device, &iec))) {
+    if ((err = scsiFetchIECmpage(device, &iec, modese_len))) {
         QUIETON(con);
         pout("unable to fetch IEC (SMART) mode page [%s]\n", 
              scsiErrString(err));
         QUIETOFF(con);
         return 1;
-    }
+    } else
+        modese_len = iec.modese_len;
+
     if ((err = scsiSetExceptionControlAndWarning(device, 0, &iec))) {
         QUIETON(con);
         pout("unable to disable Exception control and warning [%s]\n",
@@ -608,11 +826,13 @@ static int scsiSmartDisable(int device)
         return 1;
     }
     /* Need to refetch 'iec' since could be modified by previous call */
-    if ((err = scsiFetchIECmpage(device, &iec))) {
+    if ((err = scsiFetchIECmpage(device, &iec, modese_len))) {
         pout("unable to fetch IEC (SMART) mode page [%s]\n", 
              scsiErrString(err));
         return 1;
-    }
+    } else
+        modese_len = iec.modese_len;
+
     pout("Informational Exceptions (SMART) %s\n",
          scsi_IsExceptionControlEnabled(&iec) ? "enabled" : "disabled");
     pout("Temperature warning %s\n",
@@ -638,40 +858,13 @@ void scsiPrintTemp(int device)
         pout("Drive Trip Temperature:        %d C\n", trip);
 }
 
-// Compares failure type to policy in effect, and either exits or
-// simply returns to the calling routine.
-static void failuretest(int type, int returnvalue)
-{
-    // If this is an error in an "optional" SMART command
-    if (type == OPTIONAL_CMD) {
-        if (con->conservative) {
-            pout("An optional SMART command has failed: exiting.\n"
-                 "To continue, set the tolerance level to something other "
-                 "than 'conservative'\n");
-            exit(returnvalue);
-        }
-        return;
-    }
-    // If this is an error in a "mandatory" SMART command
-    if (type==MANDATORY_CMD) {
-        if (con->permissive)
-            return;
-        pout("A mandatory SMART command has failed: exiting. To continue, "
-             "use the -T option to set the tolerance level to 'permissive'\n");
-        exit(returnvalue);
-    }
-    pout("Smartctl internal error in failuretest(type=%d). Please contact "
-         "%s\n",type,PROJECTHOME);
-    exit(returnvalue|FAILCMD);
-}
-
 /* Main entry point used by smartctl command. Return 0 for success */
-int scsiPrintMain(const char *dev_name, int fd)
+int scsiPrintMain(int fd)
 {
     int checkedSupportedLogPages = 0;
     UINT8 peripheral_type = 0;
-    int returnval=0;
-    int res;
+    int returnval = 0;
+    int res, durationSec;
 
     if (scsiGetDriveInfo(fd, &peripheral_type, con->driveinfo)) {
         failuretest(MANDATORY_CMD, returnval |= FAILID);
@@ -686,7 +879,21 @@ int scsiPrintMain(const char *dev_name, int fd)
         if (scsiSmartDisable(fd))
             failuretest(MANDATORY_CMD,returnval |= FAILSMART);
     }
-
+    
+    if (con->smartautosaveenable) {
+      if (scsiSetControlGLTSD(fd, 0, modese_len)) {
+	pout("Enable autosave (clear GLTSD bit) failed\n");
+	failuretest(OPTIONAL_CMD,returnval |= FAILSMART);
+      }
+    }
+    
+    if (con->smartautosavedisable) {
+      if (scsiSetControlGLTSD(fd, 1, modese_len)) {
+	pout("Disable autosave (set GLTSD bit) failed\n");
+	failuretest(OPTIONAL_CMD,returnval |= FAILSMART);
+      }
+    }
+    
     if (con->checksmart) {
         scsiGetSupportedLogPages(fd);
         checkedSupportedLogPages = 1;
@@ -712,8 +919,18 @@ int scsiPrintMain(const char *dev_name, int fd)
                 scsiGetStartStopData(fd);
         }
     }   
-    if (con->smarterrorlog)
+    if (con->smartvendorattrib) {
+        if (gSeagateCacheLPage && (SCSI_PT_DIRECT_ACCESS == peripheral_type))
+            scsiPrintSeagateCacheLPage(fd);
+        if (gSeagateFactoryLPage && (SCSI_PT_DIRECT_ACCESS == peripheral_type))
+            scsiPrintSeagateFactoryLPage(fd);
+    }
+    if (con->smarterrorlog) {
         scsiPrintErrorCounterLog(fd);
+        if (1 == scsiFetchControlGLTSD(fd, modese_len, 1))
+            pout("\n[GLTSD (Global Logging Target Save Disable) set. "
+                 "Enable Save with '-S on']\n");
+    }
     if (con->smartselftestlog) {
         if (! checkedSupportedLogPages)
             scsiGetSupportedLogPages(fd);
@@ -755,6 +972,15 @@ int scsiPrintMain(const char *dev_name, int fd)
             return returnval;
         }
         pout("Extended Background Self Test has begun\n");
+        if ((0 == scsiFetchExtendedSelfTestTime(fd, &durationSec, 
+                        modese_len)) && (durationSec > 0)) {
+            time_t t = time(NULL);
+
+            t += durationSec;
+            pout("Please wait %d minutes for test to complete.\n", 
+                 durationSec / 60);
+            pout("Estimated completion time: %s\n", ctime(&t));
+        }
         pout("Use smartctl -X to abort test\n");        
     }
     if (con->smartextendcapselftest) {
