@@ -22,15 +22,15 @@
  *
  */
 
+#include "config.h"
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/ioctl.h>
-#include <linux/hdreg.h>
 #include <sys/fcntl.h>
 #include <sys/types.h>
 #include <string.h>
 #include <stdarg.h>
-#ifdef HAVE_GETOPT_H
+#ifdef HAVE_GETOPT_LONG
 #include <getopt.h>
 #endif
 #include "atacmds.h"
@@ -42,20 +42,22 @@
 #include "smartctl.h"
 #include "utility.h"
 
-extern const char *atacmdnames_c_cvsid, *atacmds_c_cvsid, *ataprint_c_cvsid, *escalade_c_cvsid, *knowndrives_c_cvsid, *scsicmds_c_cvsid, *scsiprint_c_cvsid, *utility_c_cvsid; 
-const char* smartctl_c_cvsid="$Id: smartctl.c,v 1.87 2003/08/18 07:10:41 ballen4705 Exp $"
-ATACMDS_H_CVSID ATAPRINT_H_CVSID EXTERN_H_CVSID KNOWNDRIVES_H_CVSID SCSICMDS_H_CVSID SCSIPRINT_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
+extern const char *atacmdnames_c_cvsid, *atacmds_c_cvsid, *ataprint_c_cvsid, *knowndrives_c_cvsid, *os_XXXX_c_cvsid, *scsicmds_c_cvsid, *scsiprint_c_cvsid, *utility_c_cvsid; 
+const char* smartctl_c_cvsid="$Id: smartctl.c,v 1.106 2003/11/26 10:37:00 dpgilbert Exp $"
+ATACMDS_H_CVSID ATAPRINT_H_CVSID CONFIG_H_CVSID EXTERN_H_CVSID KNOWNDRIVES_H_CVSID SCSICMDS_H_CVSID SCSIPRINT_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
 
 // This is a block containing all the "control variables".  We declare
 // this globally in this file, and externally in other files.
 smartmonctrl *con=NULL;
 
+// to hold onto exit code for atexit routine
+extern int exitstatus;
+
 // Track memory use
-long long bytes=0;
+extern long long bytes;
 
 void printslogan(){
-  pout("smartctl version %d.%d-%d Copyright (C) 2002-3 Bruce Allen\n",
-      (int)RELEASE_MAJOR, (int)RELEASE_MINOR, (int)SMARTMONTOOLS_VERSION);
+  pout("smartctl version %s Copyright (C) 2002-3 Bruce Allen\n", PACKAGE_VERSION);
   pout("Home page is %s\n\n",PROJECTHOME);
   return;
 }
@@ -74,10 +76,10 @@ void printcopy(){
   pout("%s",out);
   printone(out,ataprint_c_cvsid);
   pout("%s",out);
-  printone(out,escalade_c_cvsid);
-  pout("%s",out);
   printone(out,knowndrives_c_cvsid);
   pout("%s",out);
+  printone(out,os_XXXX_c_cvsid);
+  pout("%s",out);
   printone(out,scsicmds_c_cvsid);
   pout("%s",out);
   printone(out,scsiprint_c_cvsid);
@@ -86,6 +88,15 @@ void printcopy(){
   pout("%s",out);
   printone(out,utility_c_cvsid);
   pout("%s",out);
+  pout("\nsmartctl build configured on " SMARTMONTOOLS_CONFIGURE_DATE "\n");
+  pout("smartctl configure arguments: " SMARTMONTOOLS_CONFIGURE_ARGS "\n");
+
+
+  return;
+}
+
+void UsageSummary(){
+  pout("\nUse smartctl -h to get a usage summary\n\n");
   return;
 }
 
@@ -95,7 +106,7 @@ void Usage (void){
   printf("============================================ SHOW INFORMATION OPTIONS =====\n\n");
 #ifdef HAVE_GETOPT_LONG
   printf(
-"  -h, -?, --help, --usage\n"
+"  -h, --help, --usage\n"
 "         Display this help and exit\n\n"
 "  -V, --version, --copyright, --license\n"
 "         Print license, copyright, and version information and exit\n\n"
@@ -106,7 +117,7 @@ void Usage (void){
   );
 #else
   printf(
-"  -h, -?    Display this help and exit\n"
+"  -h        Display this help and exit\n"
 "  -V        Print license, copyright, and version information\n"
 "  -i        Show identity information for device\n"
 "  -a        Show all SMART information for device\n\n"
@@ -120,7 +131,7 @@ void Usage (void){
 "  -d TYPE, --device=TYPE\n"
 "         Specify device type to one of: ata, scsi, 3ware,N\n\n"
 "  -T TYPE, --tolerance=TYPE                                           (ATA)\n"
-"         Set tolerance to one of: normal, conservative, permissive\n\n"
+"         Tolerance: normal, conservative, permissive, verypermissive\n\n"
 "  -b TYPE, --badsum=TYPE                                              (ATA)\n"
 "         Set action on bad checksum to one of: warn, exit, ignore\n\n"
 "  -r TYPE, --report=TYPE\n"
@@ -130,7 +141,7 @@ void Usage (void){
   printf(
 "  -q TYPE   Set smartctl quiet mode to one of: errorsonly, silent     (ATA)\n"
 "  -d TYPE   Specify device type to one of: ata, scsi, 3ware,N\n"
-"  -T TYPE   Set tolerance to one of: normal, conservative, permissive (ATA)\n"
+"  -T TYPE   Tolerance: normal, conservative,permissive,verypermissive (ATA\n"
 "  -b TYPE   Set action on bad checksum to one of: warn, exit, ignore  (ATA)\n"
 "  -r TYPE   Report transactions (see man page)\n\n"
   );
@@ -166,7 +177,7 @@ void Usage (void){
 "  -v N,OPTION , --vendorattribute=N,OPTION                            (ATA)\n"
 "        Set display OPTION for vendor Attribute N (see man page)\n\n"
 "  -F TYPE, --firmwarebug=TYPE                                         (ATA)\n"
-"        Use firmware bug workaround. Type is one of: none, samsung\n\n"
+"        Use firmware bug workaround: none, samsung, samsung2\n\n"
 "  -P TYPE, --presets=TYPE                                             (ATA)\n"
 "        Drive-specific presets: use, ignore, show, showall\n\n"
   );
@@ -177,7 +188,7 @@ void Usage (void){
 "  -A        Show device SMART vendor-specific Attributes and values    (ATA)\n"
 "  -l TYPE   Show device log. Type is one of: error, selftest, directory\n"
 "  -v N,OPT  Set display OPTion for vendor Attribute N (see man page)   (ATA)\n"
-"  -F TYPE   Use firmware bug workaround. Type is one of: none, samsung (ATA)\n"
+"  -F TYPE   Use firmware bug workaround: none, samsung, samsung2       (ATA)\n"
 "  -P TYPE   Drive-specific presets: use, ignore, show, showall         (ATA)\n\n"
   );
 #endif
@@ -186,7 +197,7 @@ void Usage (void){
   printf(
 "  -t TEST, --test=TEST\n"
 #if DEVELOP_SELECTIVE_SELF_TEST
-"        Run test.  TEST is: offline, short, long, conveyance, selective\n\n"
+"        Run test.  TEST is: offline, short, long, conveyance, selective,M-N\n\n"
 #else
 "        Run test.  TEST is: offline, short, long, conveyance\n\n"
 #endif
@@ -204,7 +215,7 @@ void Usage (void){
 #else
   printf(
 #if DEVELOP_SELECTIVE_SELF_TEST
-"  -t TEST   Run test.  TEST is: offline, short, long, conveyance, selective\n"
+"  -t TEST   Run test.  TEST is: offline, short, long, conveyance, selective,M-N\n"
 "  -C        With -t, do test in captive mode (short/long/conveyance/selective)\n"
 #else
 "  -t TEST   Run test.  TEST is: offline, short, long, conveyance\n"
@@ -247,7 +258,7 @@ const char *getvalidarglist(char opt) {
   case 'd':
     return "ata, scsi, 3ware,N";
   case 'T':
-    return "normal, conservative, permissive";
+    return "normal, conservative, permissive, verypermissive";
   case 'b':
     return "warn, exit, ignore";
   case 'r':
@@ -262,12 +273,12 @@ const char *getvalidarglist(char opt) {
     return "use, ignore, show, showall";
   case 't':
 #if DEVELOP_SELECTIVE_SELF_TEST
-    return "offline, short, long, conveyance, selective";
+    return "offline, short, long, conveyance, selective,M-N";
 #else
     return "offline, short, long, conveyance";
 #endif
   case 'F':
-    return "none, samsung";
+    return "none, samsung, samsung2";
   case 'v':
   default:
     return NULL;
@@ -419,13 +430,14 @@ void ParseOpts (int argc, char** argv){
     case 'T':
       if (!strcmp(optarg,"normal")) {
         con->conservative = FALSE;
-        con->permissive   = FALSE;
+        con->permissive   = 0;
       } else if (!strcmp(optarg,"conservative")) {
         con->conservative = TRUE;
-        con->permissive   = FALSE;
       } else if (!strcmp(optarg,"permissive")) {
-        con->permissive   = TRUE;
-        con->conservative = FALSE;
+        if (con->permissive<0xff)
+	  con->permissive++;
+      } else if (!strcmp(optarg,"verypermissive")) {
+        con->permissive=0xff;
       } else {
         badarg = TRUE;
       }
@@ -455,7 +467,7 @@ void ParseOpts (int argc, char** argv){
           con->veryquietmode = FALSE;
           pout("Can't allocate memory to copy argument to -r option"
                " - exiting\n");
-          exit(FAILCMD);
+          EXIT(FAILCMD);
         }
         if (split_report_arg(s, &i)) {
           badarg = TRUE;
@@ -512,6 +524,8 @@ void ParseOpts (int argc, char** argv){
         con->fixfirmwarebug = FIX_NONE;
       } else if (!strcmp(optarg,"samsung")) {
         con->fixfirmwarebug = FIX_SAMSUNG;
+      } else if (!strcmp(optarg,"samsung2")) {
+        con->fixfirmwarebug = FIX_SAMSUNG2;
       } else {
         badarg = TRUE;
       }
@@ -552,16 +566,16 @@ void ParseOpts (int argc, char** argv){
         printslogan();
         if (!(s = create_vendor_attribute_arg_list())) {
           pout("Insufficient memory to construct argument list\n");
-          exit(FAILCMD);
+          EXIT(FAILCMD);
         }
         pout("The valid arguments to -v are:\n\thelp\n%s\n", s);
         free(s);
-        exit(0);
+        EXIT(0);
       }
       charp=con->attributedefs;
       if (!charp){
 	pout("Fatal internal error in ParseOpts()\n");
-	exit(FAILCMD);
+	EXIT(FAILCMD);
       }
       if (parse_attribute_def(optarg, &charp))
 	badarg = TRUE;
@@ -575,7 +589,7 @@ void ParseOpts (int argc, char** argv){
         con->showpresets = TRUE;
       } else if (!strcmp(optarg, "showall")) {
         showallpresets();
-        exit(0);
+        EXIT(0);
       } else {
         badarg = TRUE;
       }
@@ -594,9 +608,30 @@ void ParseOpts (int argc, char** argv){
         con->smartconveyanceselftest = TRUE;
         con->testcase            = CONVEYANCE_SELF_TEST;
 #if DEVELOP_SELECTIVE_SELF_TEST
-      } else if (!strcmp(optarg,"selective")) {
-        con->smartselectiveselftest = TRUE;
-        con->testcase            = SELECTIVE_SELF_TEST;
+      } else if (!strncmp(optarg,"selective",strlen("selective"))) {
+        unsigned long long start, stop;
+
+        if (split_selective_arg(optarg, &start, &stop)) {
+          badarg = TRUE;
+        } else {
+          if (con->smartselectivenumspans >= 5 || start > stop) {
+            con->veryquietmode=FALSE;
+            printslogan();
+            if (start > stop) {
+              pout("ERROR: Start LBA > ending LBA in argument \"%s\"\n",
+                optarg);
+            } else {
+              pout("ERROR: No more than five selective self-test spans may be"
+                " defined\n");
+            }
+	    UsageSummary();
+            EXIT(FAILCMD);
+          }
+          con->smartselectivespan[con->smartselectivenumspans][0] = start;
+          con->smartselectivespan[con->smartselectivenumspans][1] = stop;
+          con->smartselectivenumspans++;
+          con->testcase            = SELECTIVE_SELF_TEST;
+        }
 #endif
       } else {
         badarg = TRUE;
@@ -610,6 +645,11 @@ void ParseOpts (int argc, char** argv){
       con->testcase           = ABORT_SELF_TEST;
       break;
     case 'h':
+      con->veryquietmode=FALSE;
+      printslogan();
+      Usage();
+      EXIT(0);	
+      break;
     case '?':
     default:
       con->veryquietmode=FALSE;
@@ -625,22 +665,24 @@ void ParseOpts (int argc, char** argv){
           printvalidarglistmessage(optopt);
 	} else
 	  pout("=======> UNRECOGNIZED OPTION: %s <=======\n",arg+2);
-	pout("\nUse smartctl --help to get a usage summary\n\n");
-	exit(FAILCMD);
+	UsageSummary();
+	EXIT(FAILCMD);
       }
 #endif
       if (optopt) {
-        // Iff optopt holds a valid option then argument must be missing.
+        // Iff optopt holds a valid option then argument must be
+        // missing.  Note (BA) this logic seems to fail using Solaris
+        // getopt!
         if (strchr(shortopts, optopt) != NULL) {
           pout("=======> ARGUMENT REQUIRED FOR OPTION: %c <=======\n", optopt);
           printvalidarglistmessage(optopt);
         } else
 	  pout("=======> UNRECOGNIZED OPTION: %c <=======\n",optopt);
-	pout("\nUse smartctl -h to get a usage summary\n\n");
-	exit(FAILCMD);
+	UsageSummary();
+	EXIT(FAILCMD);
       }
       Usage();
-      exit(0);	
+      EXIT(0);	
     } // closes switch statement to process command-line options
     
     // Check to see if option had an unrecognized or incorrect argument.
@@ -651,8 +693,8 @@ void ParseOpts (int argc, char** argv){
       // a clean way to do it.
       pout("=======> INVALID ARGUMENT TO -%c: %s <======= \n", optchar, optarg);
       printvalidarglistmessage(optchar);
-      pout("\nUse smartctl -h to get a usage summary\n\n");
-      exit(FAILCMD);
+      UsageSummary();
+      EXIT(FAILCMD);
     }
   }
   // At this point we have processed all command-line options.
@@ -667,8 +709,8 @@ void ParseOpts (int argc, char** argv){
     con->veryquietmode=FALSE;
     printslogan();
     pout("\nERROR: smartctl can only run a single test (or abort) at a time.\n");
-    pout("Use smartctl -h to get a usage summary\n\n");
-    exit(FAILCMD);
+    UsageSummary();
+    EXIT(FAILCMD);
   }
 
   // If captive option was used, change test type if appropriate.
@@ -700,8 +742,8 @@ void ParseOpts (int argc, char** argv){
   // Warn if the user has provided no device name
   if (argc-optind<1){
     pout("ERROR: smartctl requires a device name as the final command-line argument.\n\n");
-    pout("Use smartctl -h to get a usage summary\n\n");
-    exit(FAILCMD);
+    UsageSummary();
+    EXIT(FAILCMD);
   }
   
   // Warn if the user has provided more than one device name
@@ -711,8 +753,8 @@ void ParseOpts (int argc, char** argv){
     pout("You have provided %d device names:\n",argc-optind);
     for (i=0; i<argc-optind; i++)
       pout("%s\n",argv[optind+i]);
-    pout("Use smartctl -h to get a usage summary\n\n");
-    exit(FAILCMD);
+    UsageSummary();
+    EXIT(FAILCMD);
   }  
 }
 
@@ -746,7 +788,8 @@ int main (int argc, char **argv){
   int fd,retval=0;
   char *device;
   smartmonctrl control;
-  int dev_type, flags;
+  int dev_type;
+  char *mode=NULL;
 
   // define control block for external functions
   con=&control;
@@ -755,20 +798,34 @@ int main (int argc, char **argv){
   ParseOpts(argc,argv);
 
   device = argv[argc-1];
-  dev_type = guess_linux_device_type(device);
-  if (GUESS_DEVTYPE_SCSI == dev_type) {
-    flags = O_RDWR | O_NONBLOCK;
-    tryscsi = 1;
-  } else { /* treat "don't know" case as an ATA device as well */
-    flags = O_RDONLY | O_NONBLOCK;
-    tryata = 1;
-  }
 
+  if (!tryata && !tryscsi) {
+    // user has not specified device type, so guess
+    dev_type = guess_device_type(device);
+    if (GUESS_DEVTYPE_SCSI == dev_type) {
+      tryscsi = 1;
+    } else if (GUESS_DEVTYPE_ATA == dev_type)
+      tryata = 1;
+    else {
+      pout("Smartctl: please specify if this is an ATA or SCSI device with the -d option.\n");
+      UsageSummary();
+      return FAILCMD;
+    }    
+  }
+  
+  // set up mode for open() call.  SCSI case is:
+  mode="SCSI";
+
+  if (tryata)
+    mode="ATA";
+    
   // open device - SCSI devices are opened (O_RDWR | O_NONBLOCK) so the
-  // scsi generci device can be used (needs write permission for MODE 
+  // scsi generic device can be used (needs write permission for MODE 
   // SELECT command) plus O_NONBLOCK to stop open hanging if media not
-  // present (e.g. with st). 
-  fd = open(device, flags);
+  // present (e.g. with st).  Opening is retried O_RDONLY if read-only
+  // media prevents opening O_RDWR (it cannot happen for scsi generic
+  // devices, but it can for the others).
+  fd = deviceopen(device, mode);
   if (fd<0) {
     char errmsg[256];
     snprintf(errmsg,256,"Smartctl open device: %s failed",argv[argc-1]);
@@ -781,12 +838,12 @@ int main (int argc, char **argv){
   if (tryata)
     retval = ataPrintMain(fd);
   else if (tryscsi)
-    retval = scsiPrintMain(device, fd);
+    retval = scsiPrintMain(fd);
   else {
     pout("Smartctl: specify if this is an ATA or SCSI device with the -d option.\n");
-    Usage();
+    UsageSummary();
     return FAILCMD;
   }
-
+  
   return retval;
 }
