@@ -22,12 +22,16 @@
  *
  */
 
+#ifndef SMARTD_H_
+#define SMARTD_H_
+
+
 #ifndef SMARTD_H_CVSID
-#define SMARTD_H_CVSID "$Id: smartd.h,v 1.45 2003/08/19 09:14:33 ballen4705 Exp $\n"
+#define SMARTD_H_CVSID "$Id: smartd.h,v 1.61 2003/11/19 19:38:19 ballen4705 Exp $\n"
 #endif
 
 // Configuration file
-#define CONFIGFILE "/etc/smartd.conf"
+#define CONFIGFILENAME "smartd.conf"
 
 // Scan directive for configuration file
 #define SCANDIRECTIVE "DEVICESCAN"
@@ -44,8 +48,9 @@
 // default for how often SMART status is checked, in seconds
 #define CHECKTIME 1800
 
-// maximum number of ATA devices to monitor
-#define MAXATADEVICES	12
+// maximum number of ATA devices to monitor.  Under linux this should
+// not exceed 20 (/dev/hda-t).  Check against make_device_names().
+#define MAXATADEVICES	20
 
 // maximum number of SCSI devices to monitor
 #define MAXSCSIDEVICES	26
@@ -62,21 +67,6 @@
 #define MONITOR_RAWPRINT  2
 #define MONITOR_RAW       3
 
-// Exit codes
-#define EXIT_BADCMD    1   // command line did not parse
-#define EXIT_BADCONF   2   // problem reading/parsing config file
-#define EXIT_STARTUP   3   // problem forking daemon
-#define EXIT_PID       4   // problem creating pid file
-
-#define EXIT_NOMEM     8   // out of memory
-#define EXIT_CCONST    9   // we hit a compile time constant
-#define EXIT_BADCODE   10  // internal error - should NEVER happen
-
-#define EXIT_BADDEV    16  // we can't monitor this device
-#define EXIT_NODEV     17  // no devices to monitor
-
-#define EXIT_SIGNAL    254 // abort on signal
-
 // If user has requested email warning messages, then this structure
 // stores the information about them.
 typedef struct mailinfo {
@@ -176,6 +166,9 @@ typedef struct configfile_s {
   char ignorepresets;                     // Ignore database of -v options
   char showpresets;                       // Show database entry for this device
   char removable;                         // Device may disappear (not be present)
+  unsigned char selflogcount;             // total number of self-test errors
+  unsigned char notused1[3];              // for packing alignment
+  unsigned short selfloghour;             // lifetime hours of last self-test error
   char *emailcmdline;                     // Program for sending mail (or NULL)
   char *address;                          // Email addresses (or NULL)
 
@@ -187,9 +180,11 @@ typedef struct configfile_s {
   unsigned char TempPageSupported;        // has log sense temperature page (0xd)
   unsigned char Temperature;              // last recorded figure (in Celsius)
   unsigned char SuppressReport;           // minimize nuisance reports
-  
+  unsigned char modese_len;               // mode sense/select cmd len: 0 (don't
+                                          // know yet) 6 or 10
+  unsigned char notused2[3];              // for packing alignment
+
   // ATA ONLY FROM HERE ON TO THE END
-  unsigned char selflogcount;             // Total number of self-test errors
   int ataerrorcount;                      // Total number of ATA errors
   
   // following NMONITOR items each point to 32 bytes, in the form of
@@ -224,6 +219,9 @@ typedef struct changedattribute_s {
 
 // Declare our own printing functions. Doing this provides error
 // messages if the argument number/types don't match the format.
+#ifndef __GNUC__
+#define __attribute__(x)      /* nothing */
+#endif
 void PrintOut(int priority,char *fmt, ...) __attribute__ ((format(printf, 2, 3)));
 
 void PrintAndMail(cfgfile *cfg, int which, int priority, char *fmt, ...) __attribute__ ((format(printf, 4, 5)));   
@@ -238,3 +236,22 @@ export NJAMD_DUMP_LEAKS_ON_EXIT=3;
 export NJAMD_TRACE_LIBS=1
 
 */
+
+// Number of seconds to allow for registering a SCSI device. If this
+// time expires without sucess or failure, then treat it as failure.
+// Set to 0 to eliminate this timeout feature from the code
+// (equivalent to an infinite timeout interval).
+#define SCSITIMEOUT 0
+
+// This is for solaris, where signal() resets the handler to SIG_DFL
+// after the first signal is caught.
+#ifdef HAVE_SIGSET
+#define SIGNALFN sigset
+#else
+#define SIGNALFN signal
+#endif
+
+#endif
+
+#define SELFTEST_ERRORCOUNT(x) (x & 0xff)
+#define SELFTEST_ERRORHOURS(x) ((x >> 8) & 0xffff)
