@@ -33,10 +33,62 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <ctype.h>
+#include <syslog.h>
+#include <stdarg.h>
 #include "utility.h"
+#include "config.h"
 
 // Any local header files should be represented by a CVSIDX just below.
-const char* utility_c_cvsid="$Id: utility.c,v 1.15 2003/08/16 22:44:50 pjwilliams Exp $" UTILITY_H_CVSID;
+const char* utility_c_cvsid="$Id: utility.c,v 1.28 2003/11/09 20:22:21 ballen4705 Exp $" CONFIG_H_CVSID UTILITY_H_CVSID;
+
+const char * packet_types[] = {
+        "Direct-access (disk)",
+        "Sequential-access (tape)",
+        "Printer",
+        "Processor",
+        "Write-once (optical disk)",
+        "CD/DVD",
+        "Scanner",
+        "Optical memory (optical disk)",
+        "Medium changer",
+        "Communications",
+        "Graphic arts pre-press (10)",
+        "Graphic arts pre-press (11)",
+        "Array controller",
+        "Enclosure services",
+        "Reduced block command (simplified disk)",
+        "Optical card reader/writer"
+};
+
+// Whenever exit() status is EXIT_BADCODE, please print this message
+const char *reportbug="Please report this bug to the Smartmontools developers at " PACKAGE_BUGREPORT ".\n";
+
+
+// hang on to exit code, so we can make use of more generic 'atexit()'
+// functionality and still check our exit code
+int exitstatus = 0;
+
+// facility used by syslog(3).  smartctl should NEVER use this, and
+// smartd can reset it with a command-line argument
+int facility=LOG_DAEMON;
+
+// command-line argument: are we running in debug mode?.
+unsigned char debugmode = 0;
+
+
+// This value follows the peripheral device type value as defined in
+// SCSI Primary Commands, ANSI INCITS 301:1997.  It is also used in
+// the ATA standard for packet devices to define the device type.
+const char *packetdevicetype(int type){
+  if (type<0x10)
+    return packet_types[type];
+  
+  if (type<0x20)
+    return "Reserved";
+  
+  return "Unknown";
+}
+
 
 // Returns 1 if machine is big endian, else zero.  This is a run-time
 // rather than a compile-time function.  We could do it at
@@ -48,20 +100,24 @@ int isbigendian(){
   return *tmp;
 }
 
+#define DBLEN 64
+
 // Utility function prints date and time and timezone into a character
 // buffer of length>=64.  All the fuss is needed to get the right
 // timezone info (sigh).
 void dateandtimezoneepoch(char *buffer, time_t tval){
   struct tm *tmval;
   char *timezonename;
-  char datebuffer[64];
+  char datebuffer[DBLEN];
   
   // Get the time structure.  We need this to determine if we are in
   // daylight savings time or not.
   tmval=localtime(&tval);
   
   // Convert to an ASCII string, put in datebuffer
-  asctime_r(tmval, datebuffer);
+  // same as: asctime_r(tmval, datebuffer);
+  strncpy(datebuffer, asctime(tmval), DBLEN);
+  datebuffer[DBLEN-1]='\0';
   
   // Remove newline
   datebuffer[strlen(datebuffer)-1]='\0';
@@ -78,7 +134,7 @@ void dateandtimezoneepoch(char *buffer, time_t tval){
     timezonename="";
   
   // Finally put the information into the buffer as needed.
-  snprintf(buffer, 64, "%s %s", datebuffer, timezonename);
+  snprintf(buffer, DBLEN, "%s %s", datebuffer, timezonename);
   
   return;
 }
@@ -97,27 +153,38 @@ void dateandtimezone(char *buffer){
 // returns distance that it has moved ahead in the input string
 int massagecvs(char *out, const char *cvsid){
   char *copy,*filename,*date,*version;
+  int retVal=0;
   const char delimiters[] = " ,$";
 
-  // make a copy on stack, go to first token,
-  if (!(copy=strdup(cvsid)) || !(filename=strtok(copy, delimiters))) 
+  // make a copy on the heap, go to first token,
+  if (!(copy=strdup(cvsid)))
     return 0;
 
+  if (!(filename=strtok(copy, delimiters))){
+    free(copy);
+    return 0;
+  }
+
   // move to first instance of "Id:"
   while (strcmp(filename,"Id:"))
-    if (!(filename=strtok(NULL, delimiters)))
+    if (!(filename=strtok(NULL, delimiters))){
+      free(copy);
       return 0;
-
+    }
+  
   // get filename, skip "v", get version and date
   if (!(  filename=strtok(NULL, delimiters)  ) ||
       !(           strtok(NULL, delimiters)  ) ||
       !(   version=strtok(NULL, delimiters)  ) ||
-      !(      date=strtok(NULL, delimiters)  ) )
+      !(      date=strtok(NULL, delimiters)  ) ) {
+    free(copy);
     return 0;
-
-   sprintf(out,"%-13s revision: %-6s date: %-15s", filename, version, date);
-   free(copy);
-   return  (date-copy)+strlen(date);
+  }
+  
+  sprintf(out,"%-13s revision: %-6s date: %-15s", filename, version, date);
+  retVal = (date-copy)+strlen(date);
+  free(copy);
+  return  retVal;
 }
 
 // prints a single set of CVS ids
@@ -129,7 +196,7 @@ void printone(char *block, const char *cvsid){
   // check that the size of the output block is sufficient
   if (len>=CVSMAXLEN) {
     pout("CVSMAXLEN=%d must be at least %d\n",CVSMAXLEN,len+1);
-    exit(1);
+    EXIT(1);
   }
 
   // loop through the different strings
@@ -189,7 +256,7 @@ int compileregex(regex_t *compiled, const char *pattern, int cflags)
   if ((errorcode = regcomp(compiled, pattern, cflags))) {
     pout("Internal error: unable to compile regular expression %s", pattern);
     printregexwarning(errorcode, compiled);
-    pout("Please inform smartmontools developers\n");
+    pout("Please inform smartmontools developers at " PACKAGE_BUGREPORT "\n");
     return 1;
   }
   return 0;
@@ -208,7 +275,7 @@ int split_report_arg(char *s, int *i)
     char *tailptr;
 
     *s++ = '\0';
-    if (*s == '0' || !isdigit(*s))  // The integer part must be positive
+    if (*s == '0' || !isdigit((int)*s))  // The integer part must be positive
       return 1;
     errno = 0;
     *i = (int) strtol(s, &tailptr, 10);
@@ -227,7 +294,7 @@ int split_report_arg2(char *s, int *i){
   char *tailptr;
   s+=6;
 
-  if (*s=='\0' || !isdigit(*s)) { 
+  if (*s=='\0' || !isdigit((int)*s)) { 
     // What's left must be integer
     *i=-1;
     return 1;
@@ -243,70 +310,112 @@ int split_report_arg2(char *s, int *i){
   return 0;
 }
 
-// Guess device type (ata or scsi) based on device name (Linux
-// specific) SCSI device name in linux can be sd, sr, scd, st, nst,
-// osst, nosst and sg.
-static const char * lin_dev_prefix = "/dev/";
-static const char * lin_dev_ata_disk_plus = "h";
-static const char * lin_dev_ata_devfs_disk_plus = "ide/";
-static const char * lin_dev_scsi_devfs_disk_plus = "scsi/";
-static const char * lin_dev_scsi_disk_plus = "s";
-static const char * lin_dev_scsi_tape1 = "ns";
-static const char * lin_dev_scsi_tape2 = "os";
-static const char * lin_dev_scsi_tape3 = "nos";
+// Splits an argument to the -t option that is assumed to be of the form
+// "selective,%lld-%lld" (prefixes of "0" (for octal) and "0x"/"0X" (for hex)
+// are allowed).  The first long long int is assigned to *start and the second
+// to *stop.  Returns zero if successful and non-zero otherwise.
+int split_selective_arg(char *s, unsigned long long *start,
+                        unsigned long long *stop)
+{
+  char *tailptr;
 
-int guess_linux_device_type(const char * dev_name) {
-  int len;
-  int dev_prefix_len = strlen(lin_dev_prefix);
-  
-  // if dev_name null, or string length zero
-  if (!dev_name || !(len = strlen(dev_name)))
-    return GUESS_DEVTYPE_DONT_KNOW;
-  
-  // Remove the leading /dev/... if it's there
-  if (!strncmp(lin_dev_prefix, dev_name, dev_prefix_len)) {
-    if (len <= dev_prefix_len)
-      // if nothing else in the string, unrecognized
-      return GUESS_DEVTYPE_DONT_KNOW;
-    // else advance pointer to following characters
-    dev_name += dev_prefix_len;
+  if (!(s = strchr(s, ',')))
+    return 1;
+  if (!isdigit((int)(*++s)))
+    return 1;
+  errno = 0;
+  // Last argument to strtoull (the base) is 0 meaning that decimal is assumed
+  // unless prefixes of "0" (for octal) or "0x"/"0X" (for hex) are used.
+  *start = strtoull(s, &tailptr, 0);
+  s = tailptr;
+  if (errno || *s++ != '-')
+    return 1;
+  *stop = strtoull(s, &tailptr, 0);
+  if (errno || *tailptr != '\0')
+    return 1;
+  return 0;
+}
+
+long long bytes = 0;
+// Helps debugging.  If the second argument is non-negative, then
+// decrement bytes by that amount.  Else decrement bytes by (one plus)
+// length of null terminated string.
+void *FreeNonZero(void *address, int size, int line, char* file){
+  if (address) {
+    if (size<0)
+      bytes-=1+strlen(address);
+    else
+      bytes-=size;
+    return CheckFree(address, line,file);
+  }
+  return NULL;
+}
+
+// To help with memory checking.  Use when it is known that address is
+// NOT null.
+void *CheckFree(void *address, int whatline,char* file){
+  if (address){
+    free(address);
+    return NULL;
   }
   
-  // form /dev/h* or h*
-  if (!strncmp(lin_dev_ata_disk_plus, dev_name,
-	       strlen(lin_dev_ata_disk_plus)))
-    return GUESS_DEVTYPE_ATA;
-  
-  // form /dev/ide/* or ide/*
-  if (!strncmp(lin_dev_ata_devfs_disk_plus, dev_name,
-	       strlen(lin_dev_ata_devfs_disk_plus)))
-    return GUESS_DEVTYPE_ATA;
+  PrintOut(LOG_CRIT, "Internal error in CheckFree() at line %d of file %s\n%s", 
+	   whatline, file, reportbug);
+  EXIT(EXIT_BADCODE);
+}
+
+
+// A custom version of strdup() that keeps track of how much memory is
+// being allocated. If mustexist is set, it also throws an error if we
+// try to duplicate a NULL string.
+char *CustomStrDup(char *ptr, int mustexist, int whatline, char* file){
+  char *tmp;
 
-  // form /dev/s* or s*
-  if (!strncmp(lin_dev_scsi_disk_plus, dev_name,
-	       strlen(lin_dev_scsi_disk_plus)))
-    return GUESS_DEVTYPE_SCSI;
+  // report error if ptr is NULL and mustexist is set
+  if (ptr==NULL){
+    if (mustexist) {
+      PrintOut(LOG_CRIT, "Internal error in CustomStrDup() at line %d of file %s\n%s", 
+	       whatline, file, reportbug);
+      EXIT(EXIT_BADCODE);
+    }
+    else
+      return NULL;
+  }
 
-  // form /dev/scsi/* or scsi/*
-  if (!strncmp(lin_dev_scsi_devfs_disk_plus, dev_name,
-	       strlen(lin_dev_scsi_devfs_disk_plus)))
-    return GUESS_DEVTYPE_SCSI;
+  // make a copy of the string...
+  tmp=strdup(ptr);
   
-  // form /dev/ns* or ns*
-  if (!strncmp(lin_dev_scsi_tape1, dev_name,
-	       strlen(lin_dev_scsi_tape1)))
-    return GUESS_DEVTYPE_SCSI;
+  if (!tmp) {
+    PrintOut(LOG_CRIT, "No memory to duplicate string %s\n", ptr);
+    EXIT(EXIT_NOMEM);
+  }
   
-  // form /dev/os* or os*
-  if (!strncmp(lin_dev_scsi_tape2, dev_name,
-	       strlen(lin_dev_scsi_tape2)))
-    return GUESS_DEVTYPE_SCSI;
+  // and track memory usage
+  bytes+=1+strlen(ptr);
   
-  // form /dev/nos* or nos*
-  if (!strncmp(lin_dev_scsi_tape3, dev_name,
-	       strlen(lin_dev_scsi_tape3)))
-    return GUESS_DEVTYPE_SCSI;
-  
-  // we failed to recognize any of the forms
-  return GUESS_DEVTYPE_DONT_KNOW;
+  return tmp;
+}
+
+// This function prints either to stdout or to the syslog as needed
+
+// [From GLIBC Manual: Since the prototype doesn't specify types for
+// optional arguments, in a call to a variadic function the default
+// argument promotions are performed on the optional argument
+// values. This means the objects of type char or short int (whether
+// signed or not) are promoted to either int or unsigned int, as
+// appropriate.]
+extern int facility;
+void PrintOut(int priority,char *fmt, ...){
+  va_list ap;
+  // initialize variable argument list 
+  va_start(ap,fmt);
+  if (debugmode) 
+    vprintf(fmt,ap);
+  else {
+    openlog("smartd",LOG_PID, facility);
+    vsyslog(priority,fmt,ap);
+    closelog();
+  }
+  va_end(ap);
+  return;
 }
